from __future__ import annotations

import os
import sys
import textwrap
from functools import reduce
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest
from docutils.core import publish_string
from docutils.utils import SystemMessage
from flow.record import RecordDescriptor

from dissect.target.exceptions import PluginError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import EmptyRecord, create_extended_descriptor
from dissect.target.plugin import (
    FunctionDescriptor,
    FunctionDescriptorLookup,
    InternalNamespacePlugin,
    InternalPlugin,
    NamespacePlugin,
    OSPlugin,
    Plugin,
    PluginDescriptor,
    PluginDescriptorLookup,
    PluginRegistry,
    _find_py_files,
    _save_plugin_import_failure,
    alias,
    environment_variable_paths,
    export,
    find_functions,
    find_functions_by_record_field_type,
    functions,
    get_external_module_paths,
    load_modules_from_paths,
    lookup,
    plugins,
)
from dissect.target.plugins.apps.other.env import EnvironmentFilePlugin
from dissect.target.plugins.general.users import UsersPlugin
from dissect.target.plugins.os.default._os import DefaultOSPlugin
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.plugins.os.unix.linux.debian._os import DebianPlugin
from dissect.target.plugins.os.unix.linux.fortios._os import FortiOSPlugin
from dissect.target.plugins.os.windows._os import WindowsPlugin
from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Iterator

    from flow.record import Record
    from pytest_benchmark.fixture import BenchmarkFixture


def test_save_plugin_import_failure() -> None:
    test_trace = ["test-trace"]
    test_module_name = "test-module"

    with (
        patch("traceback.format_exception", Mock(return_value=test_trace)),
        patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry) as mock_plugins,
    ):
        _save_plugin_import_failure(test_module_name)

        assert len(mock_plugins.__failed__) == 1
        assert mock_plugins.__failed__[0].module == test_module_name
        assert mock_plugins.__failed__[0].stacktrace == test_trace


@pytest.mark.parametrize(
    ("env_value", "expected_output"),
    [
        (None, []),
        ("", []),
        (":", [Path(), Path()]),
    ],
)
def test_load_environment_variable(env_value: str | None, expected_output: list[Path]) -> None:
    with patch.object(os, "environ", {"DISSECT_PLUGINS": env_value}):
        assert environment_variable_paths() == expected_output


def test_load_module_paths() -> None:
    assert get_external_module_paths([Path(), Path()]) == [Path()]


def test_load_paths_with_env() -> None:
    with patch.object(os, "environ", {"DISSECT_PLUGINS": ":"}):
        assert get_external_module_paths([Path(), Path()]) == [Path()]


def test_load_environment_variable_empty_string() -> None:
    with patch("dissect.target.plugin._find_py_files") as mocked_find_py_files:
        load_modules_from_paths([])
        mocked_find_py_files.assert_not_called()


def test_load_environment_variable_comma_seperated_string() -> None:
    with patch("dissect.target.plugin._find_py_files") as mocked_find_py_files:
        load_modules_from_paths([Path(), Path()])
        mocked_find_py_files.assert_called_with(Path())


def test_filter_file(tmp_path: Path) -> None:
    file = tmp_path / "hello.py"
    file.touch()

    assert list(_find_py_files(file)) == [file]

    test_file = tmp_path / "non_existent_file"
    assert list(_find_py_files(test_file)) == []

    test_file = tmp_path / "__init__.py"
    test_file.touch()
    assert list(_find_py_files(test_file)) == []


@pytest.mark.parametrize(
    ("filename", "empty_list"),
    [
        ("__init__.py", True),
        ("__pycache__/help.pyc", True),
        ("hello/test.py", False),
    ],
)
def test_filter_directory(tmp_path: Path, filename: str, empty_list: bool) -> None:
    file = tmp_path / filename
    file.parent.mkdir(parents=True, exist_ok=True)
    file.touch()

    if empty_list:
        assert list(_find_py_files(tmp_path)) == []
    else:
        assert file in list(_find_py_files(tmp_path))


@pytest.mark.parametrize(
    ("filename", "expected_module"),
    [
        ("test.py", "test"),
        ("hello_world/help.py", "hello_world.help"),
        ("path/to/file.py", "path.to.file"),
    ],
)
def test_filesystem_module_registration(tmp_path: Path, filename: str, expected_module: str) -> None:
    path = tmp_path / filename
    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch()

    load_modules_from_paths([tmp_path])

    assert expected_module in sys.modules


def test_plugin_registration(tmp_path: Path) -> None:
    code = """
        from dissect.target.plugin import Plugin, export


        class TestPlugin(Plugin):
            __register__ = False

            def check_compatible(self) -> None:
                return None

            @export(output="default")
            def hello_world(self):
                for x in self.target.fs.iterdir(""):
                    print(f"hello {x}")
    """
    (tmp_path / "plugin.py").write_text(textwrap.dedent(code))

    with patch("dissect.target.plugin.register") as mock_register:
        load_modules_from_paths([tmp_path])

        mock_register.assert_called_once()
        assert mock_register.call_args[0][0].__name__ == "TestPlugin"

    with patch("dissect.target.plugin.register") as mock_register:
        load_modules_from_paths([tmp_path / "plugin.py"])

        mock_register.assert_called_once()
        assert mock_register.call_args[0][0].__name__ == "TestPlugin"


@patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry)
def test_plugin_directory(mock_plugins: PluginRegistry, tmp_path: Path) -> None:
    code = """
        from dissect.target.plugin import Plugin, export

        class MyPlugin(Plugin):
            __namespace__ = {!r}

            @export
            def my_function(self):
                return "My function"
    """

    (tmp_path / "myplugin").mkdir()
    (tmp_path / "myplugin" / "__init__.py").write_text("")
    (tmp_path / "myplugin" / "_plugin.py").write_text(textwrap.dedent(code.format(None)))

    (tmp_path / "mypluginns").mkdir()
    (tmp_path / "mypluginns" / "__init__.py").write_text("")
    (tmp_path / "mypluginns" / "_plugin.py").write_text(textwrap.dedent(code.format("myns")))

    load_modules_from_paths([tmp_path])

    assert mock_plugins.__functions__.__regular__ == {
        "my_function": {
            "myplugin.MyPlugin": FunctionDescriptor(
                name="my_function",
                namespace=None,
                path="myplugin.my_function",
                exported=True,
                internal=False,
                findable=True,
                alias=False,
                output="default",
                method_name="my_function",
                module="myplugin._plugin",
                qualname="MyPlugin",
            )
        },
        "myns": {
            "mypluginns.MyPlugin": FunctionDescriptor(
                name="myns",
                namespace="myns",
                path="mypluginns",
                exported=True,
                internal=False,
                findable=True,
                alias=False,
                output="record",
                method_name="__call__",
                module="mypluginns._plugin",
                qualname="MyPlugin",
            )
        },
        "myns.my_function": {
            "mypluginns.MyPlugin": FunctionDescriptor(
                name="myns.my_function",
                namespace="myns",
                path="mypluginns.my_function",
                exported=True,
                internal=False,
                findable=True,
                alias=False,
                output="default",
                method_name="my_function",
                module="mypluginns._plugin",
                qualname="MyPlugin",
            )
        },
    }

    assert mock_plugins.__plugins__.__regular__ == {
        "myplugin.MyPlugin": PluginDescriptor(
            module="myplugin._plugin",
            qualname="MyPlugin",
            namespace=None,
            path="myplugin",
            findable=True,
            functions=["my_function"],
            exports=["my_function"],
        ),
        "mypluginns.MyPlugin": PluginDescriptor(
            module="mypluginns._plugin",
            qualname="MyPlugin",
            namespace="myns",
            path="mypluginns",
            findable=True,
            functions=["my_function", "__call__"],
            exports=["my_function", "__call__"],
        ),
    }


TestRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "application/test",
    [
        ("string", "value"),
    ],
)

TestRecord2 = create_extended_descriptor([UserRecordDescriptorExtension])(
    "application/test_other",
    [
        ("varint", "value"),
    ],
)


class _TestNSPlugin(NamespacePlugin):
    __namespace__ = "NS"
    __register__ = False

    @export(record=TestRecord)
    def test_all(self) -> Iterator[TestRecord]:
        # Iterate all functions of all subclasses
        yield from self.test()


class _TestSubPlugin1(_TestNSPlugin):
    __namespace__ = "t1"
    __register__ = False

    @export(record=TestRecord)
    def test(self) -> Iterator[TestRecord]:
        yield TestRecord(value="test1")


class _TestSubPlugin2(_TestNSPlugin):
    __namespace__ = "t2"
    __register__ = False

    @export(record=TestRecord)
    def test(self) -> Iterator[TestRecord]:
        yield TestRecord(value="test2")


class _TestSubPlugin3(_TestSubPlugin2):
    __namespace__ = "t3"
    __register__ = False

    # Override the test() function of t2
    @export(record=TestRecord)
    def test(self) -> Iterator[TestRecord]:
        yield TestRecord(value=self._value())

    def _value(self) -> str:
        return "test3"


class _TestSubPlugin4(_TestSubPlugin3):
    __namespace__ = "t4"
    __register__ = False

    # Do not override the test() function of t3, but change the _value function instead.
    def _value(self) -> str:
        return "test4"

    @export(record=TestRecord)
    def test_other(self) -> Iterator[TestRecord]:
        yield TestRecord(value="test4-other")

    @export(record=TestRecord)
    def test_all(self) -> Iterator[TestRecord]:
        yield TestRecord(value="overridden")


class _TestSubPlugin5(_TestNSPlugin):
    __namespace__ = "t5"
    __register__ = False

    @export(record=TestRecord2)
    def test_other(self) -> Iterator[TestRecord2]:
        yield TestRecord2(value=69)


def test_namespace_plugin(target_win: Target) -> None:
    assert "__subplugins__" in dir(_TestNSPlugin)
    # Rename the test functions to protect them from being filtered by NS

    target_win._register_plugin_functions(_TestSubPlugin1(target_win))
    target_win._register_plugin_functions(_TestSubPlugin2(target_win))
    target_win._register_plugin_functions(_TestSubPlugin3(target_win))
    target_win._register_plugin_functions(_TestSubPlugin4(target_win))
    target_win._register_plugin_functions(_TestSubPlugin5(target_win))
    target_win._register_plugin_functions(_TestNSPlugin(target_win))

    assert len(target_win.NS.__subplugins__) == 5
    assert len(target_win.NS.test.__subplugins__) == 4
    assert target_win.NS.test.__doc__ == "Return test for: t1, t2, t3, t4"
    assert [rd.name for rd in target_win.NS.test.__record__] == ["application/test"]

    assert target_win.NS.test_other.__doc__ == "Return test_other for: t4, t5"
    assert sorted([rd.name for rd in target_win.NS.test_other.__record__]) == [
        "application/test",
        "application/test_other",
    ]
    assert len(target_win.NS.test_other.__subplugins__) == 2

    assert len(list(target_win.NS.test())) == 4
    assert sorted([item.value for item in target_win.NS.test()]) == ["test1", "test2", "test3", "test4"]
    assert sorted([item.value for item in target_win.t1.test()]) == ["test1"]
    assert sorted([item.value for item in target_win.t2.test()]) == ["test2"]
    assert sorted([item.value for item in target_win.t3.test()]) == ["test3"]
    assert sorted([item.value for item in target_win.t4.test()]) == ["test4"]

    # Check whether we can access all subclass functions from the superclass
    assert sorted([item.value for item in target_win.NS.test_all()]) == ["test1", "test2", "test3", "test4"]

    # Check whether we can access the overridden function when explicitly accessing the subplugin
    assert next(target_win.t4.test_all()).value == "overridden"

    with pytest.raises(PluginError, match="Cannot merge namespace methods with different output types"):

        class _TestSubPluginFaulty(_TestNSPlugin):
            __namespace__ = "faulty"
            __register__ = False

            @export(output="yield")
            def test(self) -> Iterator[str]:
                yield "faulty"


@patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry)
def test_namespace_plugin_registration(mock_plugins: PluginRegistry) -> None:
    class _TestNSPlugin(NamespacePlugin):
        __namespace__ = "NS"

    class _TestSubPlugin1(_TestNSPlugin):
        __namespace__ = "t1"

        @export(record=TestRecord)
        def test(self) -> None: ...

    assert next(lookup("NS")).exported
    assert next(lookup("NS.test")).exported
    assert next(lookup("t1")).exported
    assert next(lookup("t1.test")).exported


@patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry)
def test_namesplace_plugin_multiple_same_module(mock_plugins: PluginRegistry) -> None:
    class NS(NamespacePlugin):
        __namespace__ = "ns"

        def check_compatible(self) -> None:
            return None

    class Foo(NS):
        __namespace__ = "foo"

        @export(output="yield")
        def baz(self) -> Iterator[str]:
            yield from ["foo"]

    class Bar(NS):
        __namespace__ = "bar"

        @export(output="yield")
        def baz(self) -> Iterator[str]:
            yield from ["bar"]

    result, _ = find_functions("*.baz")
    assert len(result) == 2
    assert sorted(desc.name for desc in result) == ["bar.baz", "foo.baz"]


@patch("dissect.target.plugin.PLUGINS", new_callable=PluginRegistry)
def test_nested_namespace_getattr(mock_plugins: PluginRegistry, target_bare: Target) -> None:
    class Foo(Plugin):
        __namespace__ = "foo"

        @export(output="yield")
        def buzz(self) -> Iterator[str]:
            yield from ["buzz"]

    class FooBar(Plugin):
        __namespace__ = "foo.bar"

        @export(output="yield")
        def bazz(self) -> Iterator[str]:
            yield from ["bazz1"]

        @export(output="yield")
        def bar(self) -> Iterator[str]:
            yield from ["bar1"]

    class FooBaz(Plugin):
        __namespace__ = "foo.baz"

        @export(output="yield")
        def bazz(self) -> Iterator[str]:
            yield from ["bazz2"]

        @export(output="yield")
        def bar(self) -> Iterator[str]:
            yield from ["bar2"]

    for plugin in [Foo, FooBar, FooBaz]:
        target_bare._register_plugin_functions(plugin(target_bare))

    assert isinstance(target_bare.foo, Foo)
    assert isinstance(target_bare.foo.bar, FooBar)
    assert isinstance(target_bare.foo.baz, FooBaz)
    assert hasattr(target_bare.foo.bar, "bazz")

    with pytest.raises(AttributeError):
        target_bare.foo.bazz()

    with pytest.raises(AttributeError):
        target_bare.foo.bar.foo()

    # Test whether we can access the plugin this way
    assert next(target_bare.foo.bar.bazz()) == "bazz1"
    assert next(target_bare.foo.bar.bar()) == "bar1"
    assert next(target_bare.foo.baz.bazz()) == "bazz2"
    assert next(target_bare.foo.baz.bar()) == "bar2"
    assert next(target_bare.foo.buzz()) == "buzz"


def test_find_plugin_function_default(target_default: Target) -> None:
    found, _ = find_functions("services", target_default)

    assert len(found) == 2
    names = [item.name for item in found]
    assert "services" in names
    assert "services" in names
    paths = [item.path for item in found]
    assert "os.unix.linux.services.services" in paths
    assert "os.windows.services.services" in paths

    found, _ = find_functions("mcafee.msc", target_default)
    assert len(found) == 1
    assert found[0].path == "apps.av.mcafee.msc"

    found, _ = find_functions("webserver.access", target_default)
    assert len(found) == 1
    assert found[0].path == "apps.webserver.webserver.access"


@pytest.mark.parametrize(
    "pattern",
    [
        ("version,ips,hostname"),
        ("ips,version,hostname"),
        ("hostname,ips,version"),
        ("users,osinfo"),
        ("osinfo,users"),
    ],
)
def test_find_plugin_function_order(target_win: Target, pattern: str) -> None:
    found = ",".join(reduce(lambda rs, el: [*rs, el.method_name], find_functions(pattern, target_win)[0], []))
    assert found == pattern


class _TestIncompatiblePlugin(Plugin):
    def check_compatible(self) -> None:
        raise UnsupportedPluginError("My incompatible plugin error")


def test_incompatible_plugin(target_bare: Target) -> None:
    with pytest.raises(UnsupportedPluginError, match="My incompatible plugin error"):
        target_bare.add_plugin(_TestIncompatiblePlugin)


# A realistic slimmed down mock of a PluginRegistry
MOCK_PLUGINS = PluginRegistry(
    __plugins__=PluginDescriptorLookup(
        __regular__={
            # Non-OS, regular
            "general.osinfo.OSInfoPlugin": PluginDescriptor(
                module="dissect.target.plugins.general.osinfo",
                qualname="OSInfoPlugin",
                namespace=None,
                path="general.osinfo",
                findable=True,
                functions=["osinfo"],
                exports=["osinfo"],
            ),
            # Non-OS, with namespace
            "apps.chat.chat.ChatPlugin": PluginDescriptor(
                module="dissect.target.plugins.apps.chat.chat",
                qualname="ChatPlugin",
                namespace="chat",
                path="apps.chat.chat",
                findable=False,
                functions=["history", "__call__"],
                exports=["history", "__call__"],
            ),
            "apps.chat.msn.MSNPlugin": PluginDescriptor(
                module="dissect.target.plugins.apps.chat.msn",
                qualname="MSNPlugin",
                namespace="msn",
                path="apps.chat.msn",
                findable=True,
                functions=["history", "__call__"],
                exports=["history", "__call__"],
            ),
            # OS, regular
            "os.windows.generic.GenericPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.windows.generic",
                qualname="GenericPlugin",
                namespace=None,
                path="os.windows.generic",
                findable=True,
                functions=["domain"],
                exports=["domain"],
            ),
            "os.unix.cronjobs.CronjobPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.cronjobs",
                qualname="CronjobPlugin",
                namespace=None,
                path="os.unix.cronjobs",
                findable=True,
                functions=["cronjobs"],
                exports=["cronjobs"],
            ),
            "os.windows.regf.runkeys.RunKeysPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.windows.regf.runkeys",
                qualname="RunKeysPlugin",
                namespace=None,
                path="os.windows.regf.runkeys",
                findable=True,
                functions=["runkeys"],
                exports=["runkeys"],
            ),
            # OS, shared name
            "os.unix.linux.services.ServicesPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.linux.services",
                qualname="ServicesPlugin",
                namespace=None,
                path="os.unix.linux.services",
                findable=True,
                functions=["initd", "services", "systemd"],
                exports=["services"],
            ),
            "os.windows.services.ServicesPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.windows.services",
                qualname="ServicesPlugin",
                namespace=None,
                path="os.windows.services",
                findable=True,
                functions=["services"],
                exports=["services"],
            ),
            # OS, with default
            "os.default.locale.LocalePlugin": PluginDescriptor(
                module="dissect.target.plugins.os.default.locale",
                qualname="LocalePlugin",
                namespace=None,
                path="os.default.locale",
                findable=True,
                functions=["timezone"],
                exports=["timezone"],
            ),
            "os.unix.locale.UnixLocalePlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.locale",
                qualname="UnixLocalePlugin",
                namespace=None,
                path="os.unix.locale",
                findable=True,
                functions=["timezone"],
                exports=["timezone"],
            ),
            "os.unix.linux.fortios.locale.FortiOSLocalePlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.linux.fortios.locale",
                qualname="FortiOSLocalePlugin",
                namespace=None,
                path="os.unix.linux.fortios.locale",
                findable=True,
                functions=["timezone"],
                exports=["timezone"],
            ),
            "os.windows.locale.WindowsLocalePlugin": PluginDescriptor(
                module="dissect.target.plugins.os.windows.locale",
                qualname="WindowsLocalePlugin",
                namespace=None,
                path="os.windows.locale",
                findable=True,
                functions=["timezone"],
                exports=["timezone"],
            ),
        },
        __os__={
            "os.default._os.DefaultOSPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.default._os",
                qualname="DefaultOSPlugin",
                namespace=None,
                path="os.default._os",
                findable=True,
                functions=["hostname", "os", "os_tree"],
                exports=["hostname", "os"],
            ),
            "os.unix._os.UnixPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix._os",
                qualname="UnixPlugin",
                namespace=None,
                path="os.unix._os",
                findable=True,
                functions=["hostname", "domain", "os", "os_tree"],
                exports=["hostname", "domain", "os"],
            ),
            "os.unix.linux._os.LinuxPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.linux._os",
                qualname="LinuxPlugin",
                namespace=None,
                path="os.unix.linux._os",
                findable=True,
                functions=["hostname", "domain", "os", "os_tree"],
                exports=["hostname", "domain", "os"],
            ),
            "os.unix.linux.fortios._os.FortiOSPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.unix.linux.fortios._os",
                qualname="FortiOSPlugin",
                namespace=None,
                path="os.unix.linux.fortios._os",
                findable=True,
                functions=["hostname", "domain", "os", "os_tree"],
                exports=["hostname", "domain", "os"],
            ),
            "os.windows._os.WindowsPlugin": PluginDescriptor(
                module="dissect.target.plugins.os.windows._os",
                qualname="WindowsPlugin",
                namespace=None,
                path="os.windows._os",
                findable=True,
                functions=["hostname", "os", "os_tree"],
                exports=["hostname", "os"],
            ),
        },
        __child__={},
    ),
    __functions__=FunctionDescriptorLookup(
        __regular__={
            # Non-OS, regular
            "osinfo": {
                "general.osinfo.OSInfoPlugin": FunctionDescriptor(
                    name="osinfo",
                    namespace=None,
                    path="general.osinfo.osinfo",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="osinfo",
                    module="dissect.target.plugins.general.osinfo",
                    qualname="OSInfoPlugin",
                )
            },
            # Non-OS, with namespace
            "chat": {
                "apps.chat.chat.ChatPlugin": FunctionDescriptor(
                    name="chat",
                    namespace="chat",
                    path="apps.chat.chat",
                    exported=True,
                    internal=False,
                    findable=False,
                    alias=False,
                    output="record",
                    method_name="__call__",
                    module="dissect.target.plugins.apps.chat.chat",
                    qualname="ChatPlugin",
                )
            },
            "chat.history": {
                "apps.chat.chat.ChatPlugin": FunctionDescriptor(
                    name="chat.history",
                    namespace="chat",
                    path="apps.chat.chat.history",
                    exported=True,
                    internal=False,
                    findable=False,
                    alias=False,
                    output="record",
                    method_name="history",
                    module="dissect.target.plugins.apps.chat.chat",
                    qualname="ChatPlugin",
                )
            },
            "msn": {
                "apps.chat.msn.MSNPlugin": FunctionDescriptor(
                    name="msn",
                    namespace="msn",
                    path="apps.chat.msn",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="__call__",
                    module="dissect.target.plugins.apps.chat.msn",
                    qualname="MSNPlugin",
                )
            },
            "msn.history": {
                "apps.chat.msn.MSNPlugin": FunctionDescriptor(
                    name="msn.history",
                    namespace="msn",
                    path="apps.chat.msn.history",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="history",
                    module="dissect.target.plugins.apps.chat.msn",
                    qualname="MSNPlugin",
                )
            },
            # OS, regular
            "domain": {
                "os.windows.generic.GenericPlugin": FunctionDescriptor(
                    name="domain",
                    namespace=None,
                    path="os.windows.generic.domain",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="domain",
                    module="dissect.target.plugins.os.windows.generic",
                    qualname="GenericPlugin",
                )
            },
            "cronjobs": {
                "os.unix.cronjobs.CronjobPlugin": FunctionDescriptor(
                    name="cronjobs",
                    namespace=None,
                    path="os.unix.cronjobs.cronjobs",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="cronjobs",
                    module="dissect.target.plugins.os.unix.cronjobs",
                    qualname="CronjobPlugin",
                )
            },
            "runkeys": {
                "os.windows.regf.runkeys.RunKeysPlugin": FunctionDescriptor(
                    name="runkeys",
                    namespace=None,
                    path="os.windows.regf.runkeys.runkeys",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="runkeys",
                    module="dissect.target.plugins.os.windows.regf.runkeys",
                    qualname="RunKeysPlugin",
                )
            },
            # OS, shared name
            "services": {
                "os.windows.services.ServicesPlugin": FunctionDescriptor(
                    name="services",
                    namespace=None,
                    path="os.windows.services.services",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="services",
                    module="dissect.target.plugins.os.windows.services",
                    qualname="ServicesPlugin",
                ),
                "os.unix.linux.services.ServicesPlugin": FunctionDescriptor(
                    name="services",
                    namespace=None,
                    path="os.unix.linux.services.services",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="record",
                    method_name="services",
                    module="dissect.target.plugins.os.unix.linux.services",
                    qualname="ServicesPlugin",
                ),
            },
            # OS, with default
            "timezone": {
                "os.default.locale.LocalePlugin": FunctionDescriptor(
                    name="timezone",
                    namespace=None,
                    path="os.default.locale.timezone",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="timezone",
                    module="dissect.target.plugins.os.default.locale",
                    qualname="LocalePlugin",
                ),
                "os.unix.locale.UnixLocalePlugin": FunctionDescriptor(
                    name="timezone",
                    namespace=None,
                    path="os.unix.locale.timezone",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="timezone",
                    module="dissect.target.plugins.os.unix.locale",
                    qualname="UnixLocalePlugin",
                ),
                "os.unix.linux.fortios.locale.FortiOSLocalePlugin": FunctionDescriptor(
                    name="timezone",
                    namespace=None,
                    path="os.unix.linux.fortios.locale.timezone",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="timezone",
                    module="dissect.target.plugins.os.unix.linux.fortios.locale",
                    qualname="FortiOSLocalePlugin",
                ),
                "os.windows.locale.WindowsLocalePlugin": FunctionDescriptor(
                    name="timezone",
                    namespace=None,
                    path="os.windows.locale.timezone",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="timezone",
                    module="dissect.target.plugins.os.windows.locale",
                    qualname="WindowsLocalePlugin",
                ),
            },
        },
        __os__={
            "hostname": {
                "os.default._os.DefaultOSPlugin": FunctionDescriptor(
                    name="hostname",
                    namespace=None,
                    path="os.default._os.hostname",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="hostname",
                    module="dissect.target.plugins.os.default._os",
                    qualname="DefaultOSPlugin",
                ),
                "os.unix._os.UnixPlugin": FunctionDescriptor(
                    name="hostname",
                    namespace=None,
                    path="os.unix._os.hostname",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="hostname",
                    module="dissect.target.plugins.os.unix._os",
                    qualname="UnixPlugin",
                ),
                "os.unix.linux._os.LinuxPlugin": FunctionDescriptor(
                    name="hostname",
                    namespace=None,
                    path="os.unix.linux._os.hostname",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="hostname",
                    module="dissect.target.plugins.os.unix.linux._os",
                    qualname="LinuxPlugin",
                ),
                "os.unix.linux.fortios._os.FortiOSPlugin": FunctionDescriptor(
                    name="hostname",
                    namespace=None,
                    path="os.unix.linux.fortios._os.hostname",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="hostname",
                    module="dissect.target.plugins.os.unix.linux.fortios._os",
                    qualname="FortiOSPlugin",
                ),
                "os.windows._os.WindowsPlugin": FunctionDescriptor(
                    name="hostname",
                    namespace=None,
                    path="os.windows._os.hostname",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="hostname",
                    module="dissect.target.plugins.os.windows._os",
                    qualname="WindowsPlugin",
                ),
            },
            "os": {
                "os.default._os.DefaultOSPlugin": FunctionDescriptor(
                    name="os",
                    namespace=None,
                    path="os.default._os.os",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="os",
                    module="dissect.target.plugins.os.default._os",
                    qualname="DefaultOSPlugin",
                ),
                "os.unix._os.UnixPlugin": FunctionDescriptor(
                    name="os",
                    namespace=None,
                    path="os.unix._os.os",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="os",
                    module="dissect.target.plugins.os.unix._os",
                    qualname="UnixPlugin",
                ),
                "os.unix.linux._os.LinuxPlugin": FunctionDescriptor(
                    name="os",
                    namespace=None,
                    path="os.unix.linux._os.os",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="os",
                    module="dissect.target.plugins.os.unix.linux._os",
                    qualname="LinuxPlugin",
                ),
                "os.unix.linux.fortios._os.FortiOSPlugin": FunctionDescriptor(
                    name="os",
                    namespace=None,
                    path="os.unix.linux.fortios._os.os",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="os",
                    module="dissect.target.plugins.os.unix.linux.fortios._os",
                    qualname="FortiOSPlugin",
                ),
                "os.windows._os.WindowsPlugin": FunctionDescriptor(
                    name="os",
                    namespace=None,
                    path="os.windows._os.os",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="os",
                    module="dissect.target.plugins.os.windows._os",
                    qualname="WindowsPlugin",
                ),
            },
            "domain": {
                "os.unix._os.UnixPlugin": FunctionDescriptor(
                    name="domain",
                    namespace=None,
                    path="os.unix._os.domain",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="domain",
                    module="dissect.target.plugins.os.unix._os",
                    qualname="UnixPlugin",
                ),
                "os.unix.linux._os.LinuxPlugin": FunctionDescriptor(
                    name="domain",
                    namespace=None,
                    path="os.unix.linux._os.domain",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="domain",
                    module="dissect.target.plugins.os.unix.linux._os",
                    qualname="LinuxPlugin",
                ),
                "os.unix.linux.fortios._os.FortiOSPlugin": FunctionDescriptor(
                    name="domain",
                    namespace=None,
                    path="os.unix.linux.fortios._os.domain",
                    exported=True,
                    internal=False,
                    findable=True,
                    alias=False,
                    output="default",
                    method_name="domain",
                    module="dissect.target.plugins.os.unix.linux.fortios._os",
                    qualname="FortiOSPlugin",
                ),
            },
        },
        __child__={},
    ),
    __ostree__={
        "os": {
            "default": {},
            "unix": {
                "linux": {
                    "fortios": {},
                }
            },
            "windows": {},
        }
    },
    __failed__=[],
)


@pytest.mark.parametrize(
    ("osfilter", "index", "expected_plugins"),
    [
        (
            None,
            "__regular__",
            [
                "dissect.target.plugins.general.osinfo",
                "dissect.target.plugins.apps.chat.chat",
                "dissect.target.plugins.apps.chat.msn",
                "dissect.target.plugins.os.windows.generic",
                "dissect.target.plugins.os.unix.cronjobs",
                "dissect.target.plugins.os.windows.regf.runkeys",
                "dissect.target.plugins.os.windows.services",
                "dissect.target.plugins.os.unix.linux.services",
                "dissect.target.plugins.os.default.locale",
                "dissect.target.plugins.os.unix.locale",
                "dissect.target.plugins.os.unix.linux.fortios.locale",
                "dissect.target.plugins.os.windows.locale",
            ],
        ),
        (
            None,
            "__os__",
            [
                "dissect.target.plugins.os.default._os",
                "dissect.target.plugins.os.unix._os",
                "dissect.target.plugins.os.unix.linux._os",
                "dissect.target.plugins.os.unix.linux.fortios._os",
                "dissect.target.plugins.os.windows._os",
            ],
        ),
        (
            "os.unix._os",
            "__regular__",
            [
                "dissect.target.plugins.general.osinfo",
                "dissect.target.plugins.apps.chat.chat",
                "dissect.target.plugins.apps.chat.msn",
                "dissect.target.plugins.os.unix.cronjobs",
                "dissect.target.plugins.os.unix.locale",
            ],
        ),
        (
            "os.unix.linux.fortios._os",
            "__regular__",
            [
                "dissect.target.plugins.general.osinfo",
                "dissect.target.plugins.apps.chat.chat",
                "dissect.target.plugins.apps.chat.msn",
                "dissect.target.plugins.os.unix.cronjobs",
                "dissect.target.plugins.os.unix.locale",
                "dissect.target.plugins.os.unix.linux.services",
                "dissect.target.plugins.os.unix.linux.fortios.locale",
            ],
        ),
        (
            "os.windows._os",
            "__regular__",
            [
                "dissect.target.plugins.general.osinfo",
                "dissect.target.plugins.apps.chat.chat",
                "dissect.target.plugins.apps.chat.msn",
                "dissect.target.plugins.os.windows.generic",
                "dissect.target.plugins.os.windows.regf.runkeys",
                "dissect.target.plugins.os.windows.services",
                "dissect.target.plugins.os.windows.locale",
            ],
        ),
        (
            "os.bar._os",
            "__regular__",
            [
                "dissect.target.plugins.general.osinfo",
                "dissect.target.plugins.apps.chat.chat",
                "dissect.target.plugins.apps.chat.msn",
            ],
        ),
        (
            "os.windows._os",
            "__os__",
            [],
        ),
    ],
)
def test_plugins(
    osfilter: str,
    index: str,
    expected_plugins: list[str],
) -> None:
    with (
        patch("dissect.target.plugin._get_plugins", return_value=MOCK_PLUGINS),
        patch("dissect.target.plugin._module_path", return_value=osfilter),
    ):
        if osfilter is not None:
            # osfilter must be a class or None
            osfilter = Mock

        plugin_descriptors = plugins(osfilter=osfilter, index=index)

        assert sorted(desc.module for desc in plugin_descriptors) == sorted(expected_plugins)


def test_plugins_default_plugin(target_default: Target) -> None:
    all_plugins = list(plugins())
    default_plugin_plugins = list(plugins(osfilter=target_default._os_plugin))

    assert default_plugin_plugins == all_plugins

    # The all_with_home is a sentinel function, which should be loaded for a
    # target with DefaultOSPlugin as OS plugin.
    sentinel_function = "all_with_home"
    has_sentinel_function = False
    for p in default_plugin_plugins:
        if sentinel_function in p.functions:
            has_sentinel_function = True
            break

    assert has_sentinel_function

    default_os_plugin_desc = plugins(osfilter=target_default._os_plugin, index="__os__")

    assert len(list(default_os_plugin_desc)) == 1


@pytest.mark.parametrize(
    ("os_plugin", "pattern", "expected_paths"),
    [
        # Direct match
        (DefaultOSPlugin, "osinfo", ["general.osinfo.osinfo"]),
        # Direct match with namespace, regardless of `findable` status
        (DefaultOSPlugin, "chat.history", ["apps.chat.chat.history"]),
        (DefaultOSPlugin, "msn.history", ["apps.chat.msn.history"]),
        # Find with tree search using wildcard
        (DefaultOSPlugin, "general.*", ["general.osinfo.osinfo"]),
        # Find with tree search using implicit wildcard
        (DefaultOSPlugin, "general.osinfo", ["general.osinfo.osinfo"]),
        # Find with exact tree match
        (DefaultOSPlugin, "general.osinfo.osinfo", ["general.osinfo.osinfo"]),
        # Find with tree search using more complicated patterns
        (DefaultOSPlugin, "general.[!o]*", []),
        (DefaultOSPlugin, "general.[!x]*", ["general.osinfo.osinfo"]),
        (DefaultOSPlugin, "general.??????.??????", ["general.osinfo.osinfo"]),
        # Namespaces do not match
        (DefaultOSPlugin, "chat.*", []),
        # Part of module paths do not match
        (DefaultOSPlugin, "generic", []),
        # OS direct match only matches within the same OS plugin
        (DefaultOSPlugin, "hostname", ["os.default._os.hostname"]),
        (WindowsPlugin, "hostname", ["os.windows._os.hostname"]),
        (LinuxPlugin, "hostname", ["os.unix.linux._os.hostname", "os.unix._os.hostname"]),
        # OS tree search only matches within the same OS plugin
        (DefaultOSPlugin, "os.windows._os.hostname", []),
        (WindowsPlugin, "os.default._os.hostname", []),
        (DefaultOSPlugin, "os.*.hostname", ["os.default._os.hostname"]),
        (WindowsPlugin, "os.*.hostname", ["os.windows._os.hostname"]),
        # # "os" hits the direct match, not the tree search
        (DefaultOSPlugin, "os", ["os.default._os.os"]),
        (WindowsPlugin, "os", ["os.windows._os.os"]),
        # Wildcard matches all regular functions and OS functions (within the same OS)
        (
            DefaultOSPlugin,
            "*",
            [
                "os.unix.linux.fortios.locale.timezone",
                "os.windows.regf.runkeys.runkeys",
                "os.unix.linux.services.services",
                "apps.chat.msn.history",
                "os.windows.generic.domain",
                "os.unix.cronjobs.cronjobs",
                "os.windows.services.services",
                "os.default.locale.timezone",
                "os.unix.locale.timezone",
                "os.windows.locale.timezone",
                "os.default._os.hostname",
                "os.default._os.os",
                "general.osinfo.osinfo",
            ],
        ),
        (
            WindowsPlugin,
            "*",
            [
                "os.windows.regf.runkeys.runkeys",
                "apps.chat.msn.history",
                "os.windows.generic.domain",
                "os.windows.services.services",
                "os.windows.locale.timezone",
                "os.windows._os.hostname",
                "os.windows._os.os",
                "general.osinfo.osinfo",
            ],
        ),
        # Some OS functions only exist on certain OS plugins
        (LinuxPlugin, "os.*.domain", ["os.unix.linux._os.domain", "os.unix._os.domain"]),
        # # Some plugins have complicated paths (i.e. LocalePlugin)
        (LinuxPlugin, "timezone", ["os.unix.locale.timezone"]),
        (FortiOSPlugin, "timezone", ["os.unix.linux.fortios.locale.timezone", "os.unix.locale.timezone"]),
        # Some plugins have overlapping names (i.e. "services")
        (FortiOSPlugin, "services", ["os.unix.linux.services.services"]),
        (WindowsPlugin, "services", ["os.windows.services.services"]),
    ],
)
def test_find_functions(os_plugin: type[OSPlugin], pattern: str, expected_paths: list[str]) -> None:
    with patch("dissect.target.plugin._get_plugins", return_value=MOCK_PLUGINS):
        target = Target()
        target._os_plugin = os_plugin

        found, _ = find_functions(pattern, target)
        assert [desc.path for desc in found] == expected_paths


def test_find_functions_windows(target_win: Target) -> None:
    found, _ = find_functions("services", target_win)

    assert len(found) == 1
    assert found[0].name == "services"
    assert found[0].path == "os.windows.services.services"


def test_find_functions_linux(target_linux: Target) -> None:
    found, _ = find_functions("services", target_linux)

    assert len(found) == 1
    assert found[0].name == "services"
    assert found[0].path == "os.unix.linux.services.services"


def test_find_functions_compatible_check(target_linux: Target) -> None:
    """Test if we correctly check for compatibility in ``find_functions`` and ``_filter_compatible``."""

    found, _ = find_functions("*", target_linux, compatibility=True)
    assert "os.unix.log.messages.syslog.syslog" not in [f"{f.path}.{f.name}" for f in found]

    with patch("dissect.target.plugins.apps.browser.chrome.ChromePlugin.check_compatible", return_value=None):
        found, _ = find_functions("*", target_linux, compatibility=True)
        functions = [f.path for f in found]
        assert "apps.browser.chrome.cookies" in functions
        assert "apps.browser.chrome.history" in functions


@pytest.mark.benchmark
def test_benchmark_functions_compatible_check(target_unix_users: Target, benchmark: BenchmarkFixture) -> None:
    """Benchmark ``_filter_compatible`` performance."""
    benchmark(lambda: find_functions("*", target_unix_users, compatibility=True))


def test_function_aliases(target_default: Target) -> None:
    """Test if alias functions are tagged as such correctly."""

    # function that is an alias should have an alias property set to True
    syslog_fd = find_functions("syslog", target_default)[0][0]
    assert syslog_fd
    assert syslog_fd.path == "os.unix.log.messages.syslog"
    assert syslog_fd.exported
    assert syslog_fd.alias

    # function that is not an alias should have an alias property set to False
    messages_fd = find_functions("messages", target_default)[0][0]
    assert messages_fd
    assert messages_fd.path == "os.unix.log.messages.messages"
    assert not messages_fd.alias


def test_function_required_arguments(target_default: Target) -> None:
    """Test if functions with required arguments are tagged as such correctly."""

    # function without any arguments should have an args property with an empty list
    syslog_fd = find_functions("syslog", target_default)[0][0]
    assert syslog_fd
    assert not syslog_fd.args

    # function with an argument should have an args property filled
    envfile_fd = find_functions("envfile", target_default)[0][0]
    assert envfile_fd
    assert envfile_fd.args == [
        (
            ("--env-path",),
            {
                "help": "path to scan environment files in",
                "required": True,
            },
        ),
        (
            ("--extension",),
            {
                "default": "env",
                "help": "extension of files to scan",
            },
        ),
    ]


def test_plugin_runtime_info() -> None:
    plugin_desc = next(p for p in plugins() if p.path == "general.users")
    assert plugin_desc.cls is UsersPlugin

    func_desc = next(p for p in functions() if p.path == "apps.other.env.envfile")
    assert func_desc.cls is EnvironmentFilePlugin
    assert func_desc.func is EnvironmentFilePlugin.envfile
    assert func_desc.record is EnvironmentFilePlugin.envfile.__record__
    assert func_desc.args == EnvironmentFilePlugin.envfile.__args__


def test_find_by_record_field_type(target_default: Target) -> None:
    assert "filesystem.walkfs.walkfs" in [desc.path for desc in find_functions_by_record_field_type("path")]
    assert "apps.other.env.envfile" in [
        desc.path for desc in find_functions_by_record_field_type("path", target_default, compatibility=True)
    ]

    with patch(
        "dissect.target.plugin.functions",
        return_value=[
            FunctionDescriptor(
                name="test",
                namespace=None,
                path="test",
                exported=True,
                internal=False,
                findable=True,
                alias=False,
                output="record",
                method_name="test",
                module="test",
                qualname="Test",
            )
        ],
    ):
        with pytest.raises(PluginError, match="An exception occurred while trying to load a plugin: test"):
            list(find_functions_by_record_field_type("path"))

        assert list(find_functions_by_record_field_type("path", ignore_load_errors=True)) == []


@pytest.mark.parametrize(
    "method_name",
    [
        "hostname",
        "ips",
        "version",
        "os",
        "architecture",
    ],
)
def test_os_plugin_property_methods(target_bare: Target, method_name: str) -> None:
    os_plugin = OSPlugin(target_bare)
    with pytest.raises(NotImplementedError):
        getattr(os_plugin, method_name)


class MockOS1(OSPlugin):
    __register__ = False

    @export(property=True)
    def hostname(self) -> str | None:
        pass

    @export(property=True)
    def ips(self) -> list[str]:
        pass

    @export(property=True)
    def version(self) -> str | None:
        pass

    @export(record=EmptyRecord)
    def users(self) -> list[Record]:
        pass

    @export(property=True)
    def os(self) -> str:
        pass

    @export(property=True)
    def architecture(self) -> str | None:
        pass


class MockOS2(OSPlugin):
    __register__ = False

    @export(property=True)
    def hostname(self) -> str | None:
        """Test docstring hostname."""

    @export(property=True)
    def ips(self) -> list[str]:
        """Test docstring ips."""

    @export(property=True)
    def version(self) -> str | None:
        """Test docstring version."""

    @export(record=EmptyRecord)
    def users(self) -> list[Record]:
        """Test docstring users."""

    @export(property=True)
    def os(self) -> str:
        """Test docstring os."""

    @export(property=True)
    def architecture(self) -> str | None:
        """Test docstring architecture."""


@pytest.mark.parametrize(
    ("subclass", "replaced"),
    [
        (MockOS1, True),
        (MockOS2, False),
    ],
)
def test_os_plugin___init_subclass__(subclass: type[OSPlugin], replaced: bool) -> None:
    exported_methods = {
        "hostname",
        "ips",
        "version",
        "users",
        "os",
        "architecture",
    }

    for method_name in exported_methods:
        os_method = getattr(OSPlugin, method_name)
        if isinstance(os_method, property):
            os_method = os_method.fget
        os_docstring = os_method.__doc__

        subclass_method = getattr(subclass, method_name)
        if isinstance(subclass_method, property):
            subclass_method = subclass_method.fget
        subclass_docstring = subclass_method.__doc__

        assert (os_docstring == subclass_docstring) is replaced
        if not replaced:
            assert subclass_docstring == f"Test docstring {method_name}."


@pytest.mark.parametrize(
    argnames=("os_plugin", "results"),
    argvalues=[
        (LinuxPlugin, ["linux", "unix"]),
        (FortiOSPlugin, ["fortios", "linux", "unix"]),
        (DebianPlugin, ["linux", "unix"]),
    ],
)
def test_os_tree(target_bare: Target, os_plugin: type[OSPlugin], results: list[str]) -> None:
    """Test if we correctly return the OS name tree."""
    target_bare._os_plugin = os_plugin
    target_bare.apply()
    os_funcs = [record.name for record in target_bare.osinfo()]

    # Ensure that os_tree is not present in os_info
    assert "os_tree" not in os_funcs
    assert target_bare.os_tree() == results


class _TestInternalPlugin(InternalPlugin):
    def test(self) -> None:
        pass


def test_internal_plugin() -> None:
    assert "test" not in _TestInternalPlugin.__exports__
    assert "test" in _TestInternalPlugin.__functions__


class _TestInternalNamespacePlugin(InternalNamespacePlugin):
    __namespace__ = "NS"

    def test(self) -> None:
        pass


def test_internal_namespace_plugin() -> None:
    assert "__subplugins__" in dir(_TestInternalNamespacePlugin)
    assert "test" not in _TestInternalNamespacePlugin.__exports__
    assert "test" in _TestInternalNamespacePlugin.__functions__


class ExampleFooPlugin(Plugin):
    """Example Foo Plugin."""

    __register__ = False

    def check_compatible(self) -> None:
        return

    @export
    @alias("bar")
    @alias(name="baz")
    def foo(self) -> Iterator[str]:
        """Yield foo!."""
        yield "foo!"


def test_plugin_alias(target_bare: Target) -> None:
    """Test ``@alias`` decorator behaviour."""
    target_bare.add_plugin(ExampleFooPlugin)
    assert target_bare.has_function("foo")
    assert target_bare.foo.__aliases__ == ["baz", "bar"]
    assert target_bare.has_function("bar")
    assert target_bare.has_function("baz")
    assert list(target_bare.foo()) == list(target_bare.bar()) == list(target_bare.baz())


@pytest.mark.parametrize(
    "descriptor",
    find_functions("*", Target(), compatibility=False, show_hidden=True)[0],
    ids=lambda d: d.path,
)
def test_exported_plugin_format(descriptor: FunctionDescriptor) -> None:
    """This test checks plugin style guide conformity for all exported plugins.

    References:
        - https://docs.dissect.tools/en/latest/contributing/style-guide.html
    """
    # Ignore DefaultOSPlugin, NamespacePlugin and OSPlugin instances
    if issubclass(descriptor.cls, (NamespacePlugin, OSPlugin)) or descriptor.cls is DefaultOSPlugin:
        return

    # Plugin method should specify what it returns
    assert descriptor.output in ["record", "yield", "default", "none"], (
        f"Invalid output_type for function {descriptor.func.__qualname__}"
    )

    annotations = None

    if hasattr(descriptor.func, "__annotations__"):
        annotations = descriptor.func.__annotations__

    elif isinstance(descriptor.func, property):
        annotations = descriptor.func.fget.__annotations__

    # Plugin method should have a return annotation
    assert annotations
    assert "return" in annotations, f"No return type annotation for function {descriptor.func.__qualname__}"

    # TODO: Check if the annotations make sense with the provided output_type

    # Plugin method should have a docstring
    method_doc_str = descriptor.func.__doc__
    assert isinstance(method_doc_str, str), f"No docstring for function {descriptor.func.__qualname__}"
    assert method_doc_str != "", f"Empty docstring for function {descriptor.func.__qualname__}"

    # The method docstring should compile to rst without warnings
    assert_valid_rst(method_doc_str)

    # The method docstring should follow our conventions
    assert_compliant_rst(method_doc_str)

    # Plugin class should have a docstring
    class_doc_str = descriptor.cls.__doc__
    assert isinstance(class_doc_str, str), f"No docstring for class {descriptor.cls.__name__}"
    assert class_doc_str != "", f"Empty docstring for class {descriptor.cls.__name__}"

    # The class docstring should compile to rst without warnings
    assert_valid_rst(class_doc_str)

    # The class docstring should follow our conventions
    assert_compliant_rst(class_doc_str)

    # Arguments of the plugin should define their type and if they are required (explicitly or implicitly).
    for arg in descriptor.args:
        names, settings = arg
        is_bool_action = settings.get("action", "") in (
            "store_true",
            "store_false",
        )

        assert names, f"No argument names for argument of function {descriptor.func.__qualname__}"
        assert sorted(names, key=len) == list(names), (
            f"Argument names {names!r} for function {descriptor.func.__qualname__} should specify short form first"
        )

        assert settings.get("default", 1) is not None, (
            f"Superfluous default of None for argument {names[0]} in function {descriptor.func.__qualname__}: "
            "default is implied as None already."
        )

        assert settings.get("help"), f"No help text for argument {names[0]} in function {descriptor.func.__qualname__}"

        dest = settings.get("dest") or names[-1].strip("-").replace("-", "_")
        assert dest in annotations, (
            f"Missing type annotation for argument {dest} in function {descriptor.func.__qualname__}"
        )

        # TODO: More strictly check type annotation, use a contains right now to also match optionals
        type_ = "bool" if is_bool_action else getattr(settings.get("type"), "__name__", "str")
        assert type_ in annotations[dest], (
            f"Invalid type annotation for argument {dest} in function {descriptor.func.__qualname__} "
            f"({annotations[dest]} instead of {type_})"
        )

        assert settings.get("type") is not str, (
            f"Superfluous type of str for argument {names[0]} in function {descriptor.func.__qualname__}: "
            "type is implied as str by default."
        )

        # Inverse checks

        if settings.get("required"):
            assert not settings.get("default"), (
                "It does not make sense to set an argument to required and have a default value"
                f"in {names[0]} in function {descriptor.func.__qualname__}"
            )

        if "required" in settings:
            assert settings.get("required"), (
                f"Superfluous required of False for argument {names[0]} in function {descriptor.func.__qualname__}: "
                "required is implied as False already."
            )

        if is_bool_action:
            assert "type" not in settings, (
                f"Type should not be set for store_true or store_false in {names[0]} in "
                f"function {descriptor.func.__qualname__}: "
                "type is implied as boolean already."
            )

            assert "default" not in settings, (
                f"Default should not be set for store_true or store_false in {names[0]} in "
                f"function {descriptor.func.__qualname__}: "
                "default is implied as opposite boolean already."
            )


def test_plugin_record_field_and_name_consistency() -> None:
    """Test that ensures that record fields with the same name across different plugins translate to the
    same normalized type, and that no two records share the same name.

    Consider the following TargetRecordDescriptors for plugins X, Y, Z, A and B::

        # Field consistency example
        RecordX = TargetRecordDescriptor("record/x", [("varint", "shared_field")])
        RecordY = TargetRecordDescriptor("record/y", [("path", "shared_field")])
        RecordZ = TargetRecordDescriptor("record/z", [("string", "shared_field")])

        # Record name consistency example
        RecordA = TargetRecordDescriptor("record/a", [("string", "field1")])
        RecordB = TargetRecordDescriptor("record/a", [("varint", "field2")])

    In the first example, ``RecordX`` will fail the test, as the field ``shared_field`` cannot be ``varint`` in
    one record also being used as ``string``/``path`` in others. ``RecordY`` and ``RecordZ`` do
    not conflict, as both translate to the same ``wildcard`` type.

    In the second example, ``RecordA`` and ``RecordB`` will fail the test because they both define the name
    ``record/a``. Duplicate record names may cause collisions in export formats or downstream processing.

    Field type normalization is based on ``FIELD_TYPES_MAP``, loosely aligned with flow.record and
    ElasticSearch type mappings.

    References:
        - https://elastic.co/guide/en/elasticsearch/reference/current/mapping-types.html
        - https://github.com/fox-it/flow.record/tree/main/flow/record/fieldtypes
        - https://github.com/JSCU-NL/dissect-elastic
    """
    seen_records: set[RecordDescriptor] = set()
    seen_field_names: set[str] = set()
    seen_field_types: dict[str, tuple[str | None, RecordDescriptor]] = {}
    seen_record_names: set[str] = set()
    inconsistencies: set[str] = set()

    FIELD_TYPES_MAP = {
        # strings
        "string": "string",
        "stringlist": "string",
        "wstring": "string",
        "path": "string",
        "uri": "string",
        "command": "string",
        "dynamic": "string",
        # ints
        "varint": "int",
        "filesize": "int",
        "uint32": "int",
        "uint16": "int",
        "float": "float",
        # ip / cidr
        "net.ipaddress": "ip",
        "net.ipnetwork": "ip_range",
        "net.ipinterface": "ip_range",
        # dates
        "datetime": "datetime",
        # other
        "boolean": "boolean",
        "bytes": "binary",
        "digest": "keyword",
    }

    for descriptor in find_functions("*", Target(), compatibility=False, show_hidden=True)[0]:
        # Test if plugin function record fields make sense and do not conflict with other records.
        if descriptor.output != "record" or not hasattr(descriptor, "record"):
            continue

        # Functions can yield a single record or a list of records.
        records = descriptor.record if isinstance(descriptor.record, list) else [descriptor.record]

        for record in records:
            assert isinstance(record, RecordDescriptor), (
                f"{record!r} of function {descriptor!r} is not of type RecordDescriptor"
            )
            if record.name != "empty":
                assert record.fields, f"{record!r} has no fields"

            if record in seen_records:
                continue

            seen_records.add(record)
            # Detect duplicate record names
            if record.name in seen_record_names:
                inconsistencies.add(f"<{record.name}> has duplicates")
            else:
                seen_record_names.add(record.name)

            for name, field in record.fields.items():
                # Make sure field names have the same type when translated. This check does not save multiple field
                # name and typenames, this is a bare-minumum check only.

                # We only care about the field type, not if it is a list of that type.
                field_typename = field.typename.replace("[]", "")

                assert field_typename in FIELD_TYPES_MAP, (
                    f"Field type {field_typename} is not mapped in FIELD_TYPES_MAP, please add it manually."
                )

                if name in seen_field_names:
                    seen_typename, seen_record = seen_field_types[name]
                    if FIELD_TYPES_MAP[seen_typename] != FIELD_TYPES_MAP[field_typename]:
                        inconsistencies.add(
                            f"<{record.name} ({field.typename!r}, '{name}')> is duplicate mismatch of <{seen_record.name} ({seen_typename!r}, '{name}')>"  # noqa: E501
                        )

                else:
                    seen_field_names.add(name)
                    seen_field_types[name] = (field_typename, record)

    if inconsistencies:
        pytest.fail(
            f"Found {len(inconsistencies)} inconsistencies in RecordDescriptors:\n" + "\n".join(inconsistencies)
        )


def assert_valid_rst(src: str) -> None:
    """Attempts to compile the given string to rst."""

    try:
        publish_string(src, settings_overrides={"halt_level": 2})

    except SystemMessage as e:
        # Limited context was provided to docutils, so some exceptions could incorrectly be raised.
        # We can assume that if the rst is truly invalid this will also be caught by `tox -e build-docs`.
        if "Unknown interpreted text role" not in str(e):
            pytest.fail(f"Invalid rst: {e}", pytrace=False)


def assert_compliant_rst(src: str) -> None:
    """Makes sure that the given rst docstring follows the project's conventions."""

    # Explicit message stating we want References instead of Resources to prevent confusion
    if "Resources:\n" in src:
        pytest.fail(f"Invalid rst: docstring contains 'Resources' instead of 'References': {src!r}", pytrace=False)

    # Generic message stating lists should start with References (assumes lists always have at least one 'http')
    if "- http" in src and "References:\n" not in src:
        pytest.fail(f"Invalid rst: docstring contains list but does not mention 'References': {src!r}", pytrace=False)

    # Make sure we use stripes instead of bullets (assumes lists always have at least one 'http')
    if "* http" in src:
        pytest.fail(f"Invalid rst: docstring contains bullet instead of dash in list: {src!r}", pytrace=False)


@pytest.mark.parametrize(
    "descriptor",
    [descriptor for descriptor in plugins() if descriptor.namespace and "." in descriptor.namespace],
    ids=lambda descriptor: descriptor.namespace,
)
def test_nested_namespace_consistency(descriptor: PluginDescriptor) -> None:
    """Test whether all parts of nested namespaces exist and that there are no conflicts with other functions."""

    parts = descriptor.namespace.split(".")
    for i in range(len(parts)):
        part = ".".join(parts[: i + 1])
        result = list(lookup(part))

        if not result:
            pytest.fail(f"Unreachable namespace {descriptor.namespace!r}, namespace {part!r} does not exist.")

        if len(result) > 1:
            conflicts = ", ".join(
                f"{desc.name} ({desc.module}.{desc.qualname})" for desc in result if desc.namespace != part
            )
            pytest.fail(f"Namespace name {descriptor.namespace!r} has conflicts with function name: {conflicts}")


@pytest.mark.parametrize(
    "descriptor",
    # Match plugin classes which are a *direct* base of NamespacePlugin only using :meth:`Plugin.__bases__`,
    # instead of using ``issubclass`` which would also yield indirectly inherited Plugin classes.
    [descriptor for descriptor in plugins() if NamespacePlugin in descriptor.cls.__bases__],
    ids=lambda descriptor: descriptor.qualname,
)
def test_namespace_class_usage(descriptor: PluginDescriptor) -> None:
    """This test checks if :class:`NamespacePlugin` usage is correct.

    :class:`NamespacePlugin` is reserved for "grouping" other plugins of the same category. See for
    example :class:`BrowserPlugin` or :class:`WebserverPlugin`.

    If you want to expose plugin functions under a shared name, e.g. ``foo.bar`` and ``foo.baz``,
    you should use :class:`Plugin` with ``Plugin.__namespace__ = "foo"`` instead.

    References:
        - https://github.com/fox-it/dissect.target/issues/1180
    """

    assert descriptor.cls.__subclasses__(), (
        f"NamespacePlugin {descriptor.module}.{descriptor.qualname} has no subclasses, are you sure you're using NamespacePlugin correctly?"  # noqa: E501
    )



from __future__ import annotations

import logging
import platform
import tarfile
from typing import TYPE_CHECKING

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.tar import GenericTarSubLoader, TarLoader
from dissect.target.plugins.os.windows._os import WindowsPlugin
from dissect.target.target import Target
from tests._utils import absolute_path
from tests.filesystems.test_tar import _mkdir

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``TarLoader`` when opening a ``Target``."""
    path = absolute_path("_data/loaders/tar/test-archive.tar")

    target = opener(path)
    assert isinstance(target._loader, TarLoader)
    assert target.path == path


def test_compressed_tar_file(caplog: pytest.LogCaptureFixture) -> None:
    """Test if we can handle a compressed tar file."""
    path = absolute_path("_data/loaders/tar/test-archive.tar.gz")

    with caplog.at_level(logging.WARNING):
        loader = loader_open(path)
        assert isinstance(loader, TarLoader)
        assert "is compressed" in caplog.text

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, GenericTarSubLoader)

    assert len(t.filesystems) == 1

    test_file = t.fs.path("test-data/test-file.txt")

    assert test_file.exists()
    assert test_file.open().read() == b"test-value\n"


def test_compressed_tar_file_with_empty_dir() -> None:
    """Test if we can handle a tar file with an empty directory."""
    path = absolute_path("_data/loaders/tar/test-archive-empty-folder.tgz")

    loader = loader_open(path)
    assert isinstance(loader, TarLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, GenericTarSubLoader)

    assert len(t.filesystems) == 1

    test_file = t.fs.path("test/test_file_with_content")
    assert test_file.exists()
    assert test_file.is_file()
    assert test_file.open().read() == b"This is a test!\n"

    empty_folder = t.fs.path("test/empty_dir")
    assert empty_folder.exists()
    assert empty_folder.is_dir()


def test_case_sensitivity_windows(tmp_path: Path) -> None:
    """Test if we correctly map a tar with Windows folder structure as a case-insensitive filesystem."""
    path = tmp_path.joinpath("target.tar.gz")
    with tarfile.open(path, "w:gz") as tf:
        _mkdir(tf, "Windows")
        _mkdir(tf, "Windows/System32")

    loader = loader_open(path)
    assert isinstance(loader, TarLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, GenericTarSubLoader)

    # Make sure the case sensitiveness is changed to False and make sure we detect the target as Windows.
    assert not t.filesystems[0].case_sensitive
    assert WindowsPlugin.detect(t)


def test_case_sensitivity_linux(tmp_path: Path) -> None:
    """Test if we correctly map a tar with Linux folder structure as a case-sensitive filesystem."""
    path = tmp_path.joinpath("target.tar.gz")
    with tarfile.open(path, "w:gz") as tf:
        _mkdir(tf, "etc")
        _mkdir(tf, "var")
        _mkdir(tf, "opt")

    loader = loader_open(path)
    assert isinstance(loader, TarLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, GenericTarSubLoader)

    assert t.filesystems[0].case_sensitive


@pytest.mark.parametrize(
    ("should_detect", "filename", "buffer"),
    [
        # regular tar file
        (True, "file.tar", ""),
        # gzip tar file
        (True, "file.tar.gz", ""),
        (True, "file.tgz", ""),
        # bzip2 tar file
        (True, "file.tar.bz2", ""),
        (True, "file.tar.bz", ""),
        (True, "file.tbz", ""),
        (True, "file.tbz2", ""),
        # xz tar file
        (True, "file.tar.xz", ""),
        (True, "file.txz", ""),
        # some things it should not detect
        (False, "file", "00010203"),
        (False, "file.zip", "504b0304"),
    ],
)
def test_detect_extension(should_detect: bool, filename: str, buffer: str, tmp_path: Path) -> None:
    """Test if we detect the given buffer as a (compressed) tar file or not."""
    tmp_tar = tmp_path.joinpath(filename)
    tmp_tar.touch()
    with tmp_tar.open("wb") as fh:
        fh.write(bytes.fromhex(buffer))
    assert TarLoader.detect(tmp_tar) == should_detect


@pytest.mark.parametrize(
    "file",
    [
        "small.tar",
        "small.tar.bz2",
        "small.tar.gz",
        "small.tar.lz",
        "small.tar.xz",
    ],
)
def test_detect_buffer(file: str, tmp_path: Path) -> None:
    """Test if we detect the given files as a (compressed) tar file or not."""

    if file == "small.tar.lz" and (platform.python_implementation() == "PyPy" or platform.system() == "Windows"):
        pytest.skip(reason="LZMA is flaky on PyPy and/or Windows")

    small_file = absolute_path(f"_data/loaders/tar/detect/{file}")

    # We rename the file to prevent detection based on file suffix.
    tmp_tar = tmp_path.joinpath(file.replace(".", "-"))
    tmp_tar.write_bytes(small_file.read_bytes())

    assert TarLoader.detect(tmp_tar)



from __future__ import annotations

import base64
import binascii
import hashlib
from pathlib import Path

from flow.record import RecordDescriptor

try:
    from asn1crypto import pem, x509

    HAS_ASN1 = True

except ImportError:
    HAS_ASN1 = False

COMMON_CERTIFICATE_FIELDS = [
    ("digest", "fingerprint"),
    ("varint", "serial_number"),
    ("string", "serial_number_hex"),
    ("datetime", "not_valid_before"),
    ("datetime", "not_valid_after"),
    ("string", "issuer_dn"),
    ("string", "subject_dn"),
    ("bytes", "pem"),
]

CertificateRecord = RecordDescriptor(
    "certificate",
    [
        *COMMON_CERTIFICATE_FIELDS,
    ],
)

# Translation layer for asn1crypto names to RFC4514 names.
# References: https://github.com/wbond/asn1crypto/blob/master/asn1crypto/x509.py @ NameType
# References: https://github.com/pyca/cryptography/blob/main/src/cryptography/x509/name.py
NAMEOID_TO_NAME = {
    "common_name": "CN",  # 2.5.4.3
    "country_name": "C",  # 2.5.4.6
    "locality_name": "L",  # 2.5.4.7
    "state_or_province_name": "ST",  # 2.5.4.8
    "street_address": "STREET",  # 2.5.4.9
    "organization_name": "O",  # 2.5.4.10
    "organizational_unit_name": "OU",  # 2.5.4.11
    "domain_component": "DC",  # 0.9.2342.192.00300.100.1.25
    "user_id": "UID",  # 0.9.2342.192.00300.100.1.1
}


def compute_pem_fingerprints(pem: str | bytes) -> tuple[str, str, str]:
    """Compute the MD5, SHA-1 and SHA-256 fingerprint hash of a x509 certificate PEM."""

    if pem is None:
        raise ValueError("No PEM provided")

    if isinstance(pem, bytes):
        pem = pem.decode()

    elif not isinstance(pem, str):
        raise TypeError("Provided PEM is not str or bytes")

    stripped_pem = pem.strip().removeprefix("-----BEGIN CERTIFICATE-----").removesuffix("-----END CERTIFICATE-----")

    try:
        der = base64.b64decode(stripped_pem)
    except binascii.Error as e:
        raise ValueError(f"Unable to parse PEM: {e!s}") from e

    md5 = hashlib.md5(der).hexdigest()
    sha1 = hashlib.sha1(der).hexdigest()
    sha256 = hashlib.sha256(der).hexdigest()

    return md5, sha1, sha256


def format_serial_number_as_hex(serial_number: int | None) -> str | None:
    """Format serial_number from integer to hex.

    Add a prefix 0 if output length is not pair, in order to be consistent with usual serial_number representation
    (navigator, openssl etc...).
    For negative number use the same representation as navigator, which differ from OpenSSL.

    For example for -1337::

        OpenSSL : Serial Number: -1337 (-0x539)
        Navigator : FA C7

    Args:
        serial_number: The serial number to format as hex.
    """
    if serial_number is None:
        return serial_number

    if serial_number > 0:
        serial_number_as_hex = f"{serial_number:x}"
        if len(serial_number_as_hex) % 2 == 1:
            serial_number_as_hex = f"0{serial_number_as_hex}"
        return serial_number_as_hex
    # Representation is always a multiple of 8 bits, we need to compute this size
    output_bin_len = 8 - (serial_number.bit_length() % 8) + serial_number.bit_length()
    return f"{serial_number & ((1 << output_bin_len) - 1):x}"


def parse_x509(file: str | bytes | Path) -> CertificateRecord:
    """Parses a PEM file. Returns a CertificateREcord. Does not parse a public key embedded in a x509 certificate."""

    if isinstance(file, str):
        content = file.encode()

    elif isinstance(file, bytes):
        content = file

    elif isinstance(file, Path) or hasattr(file, "read_bytes"):
        content = file.read_bytes()

    else:
        raise TypeError("Parameter file is not of type str, bytes or Path")

    if not HAS_ASN1:
        raise ValueError("Missing asn1crypto dependency")

    md5, _, _ = compute_pem_fingerprints(content.decode())
    _, _, der = pem.unarmor(content)
    crt = x509.Certificate.load(der)

    issuer = []
    for key, value in crt.issuer.native.items():
        issuer.append(f"{NAMEOID_TO_NAME.get(key, key)}={value}")

    subject = []
    for key, value in crt.subject.native.items():
        subject.append(f"{NAMEOID_TO_NAME.get(key, key)}={value}")

    return CertificateRecord(
        not_valid_before=crt.not_valid_before,
        not_valid_after=crt.not_valid_after,
        issuer_dn=",".join(issuer),
        subject_dn=",".join(subject),
        fingerprint=(md5, crt.sha1.hex(), crt.sha256.hex()),
        serial_number=crt.serial_number,
        serial_number_hex=format_serial_number_as_hex(crt.serial_number),
        pem=crt.dump(),
    )



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct
from dissect.util import lzxpress_huffman
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    from dissect.target.target import Target

PrefetchRecord = TargetRecordDescriptor(
    "filesystem/ntfs/prefetch",
    [
        ("datetime", "ts"),
        ("path", "filename"),
        ("path", "prefetch"),
        ("path", "linkedfile"),
        ("uint32", "runcount"),
    ],
)


CompactPrefetchRecord = TargetRecordDescriptor(
    "filesystem/ntfs/prefetch/compact",
    [
        ("datetime", "ts"),
        ("path", "filename"),
        ("path", "prefetch"),
        ("path[]", "linkedfiles"),
        ("uint32", "runcount"),
        ("datetime[]", "previousruns"),
    ],
)


prefetch_def = """
    struct PREFETCH_HEADER_DETECT {
        char signature[4];
        uint32 size;
    };

    struct PREFETCH_HEADER {
        uint32 version;
        char signature[4];
        uint32 unknown;
        uint32 size;
        char name[60];
        uint32 hash;
        uint32 flag;
    };

    struct FILE_INFORMATION_26 {
        uint32 metrics_array_offset;
        uint32 number_of_file_metrics_entries;
        uint32 trace_chain_array_offset;
        uint32 number_of_trace_chain_array_entries;
        uint32 filename_strings_offset;
        uint32 filename_strings_size;
        uint32 volumes_information_offset;
        uint32 number_of_volumes;
        uint32 volumes_information_size;
        uint32 unknown0[2];
        uint64 last_run_time;
        uint64 last_run_remains[7];
        uint64 unknown1[2];
        uint32 run_count;
        uint32 unknown2;
        uint32 unknown3;
        char unknown4[88];
    };

    struct FILE_INFORMATION_17 {
        uint32 metrics_array_offset;
        uint32 number_of_file_metrics_entries;
        uint32 trace_chain_array_offset;
        uint32 number_of_trace_chain_array_entries;
        uint32 filename_strings_offset;
        uint32 filename_strings_size;
        uint32 volumes_information_offset;
        uint32 number_of_volumes;
        uint32 volumes_information_size;
        uint32 last_run_time;
        uint32 unknown0;
        uint32 run_count;
        uint32 unknown1;
    };

    struct FILE_INFORMATION_23 {
        uint32 metrics_array_offset;
        uint32 number_of_file_metrics_entries;
        uint32 trace_chain_array_offset;
        uint32 number_of_trace_chain_array_entries;
        uint32 filename_strings_offset;
        uint32 filename_strings_size;
        uint32 volumes_information_offset;
        uint32 number_of_volumes;
        uint32 volumes_information_size;
        uint32 unknown[2];
        uint64 last_run_time;
        uint64 last_run_remains[2];
        uint32 run_count;
        uint32 unknown0;
        uint32 unknown1;
        char unknown2[80];
    };

    struct VOLUME_INFORMATION_17 {
        uint32 device_path_offset;
        uint32 device_path_number_of_characters;
        uint64 creation_time;
        uint32 serial_number;
        uint32 file_reference_offset;
        uint32 file_reference_size;
        uint32 directory_strings_array_offset;
        uint32 number_of_directory_strings;
        uint32 unknown;
    };

    struct VOLUME_INFORMATION_30 {
        uint32 device_path_offset;
        uint32 device_path_number_of_characters;
        uint64 creation_time;
        uint32 serial_number;
        uint32 file_reference_offset;
        uint32 file_reference_size;
        uint32 directory_strings_array_offset;
        uint32 number_of_directory_strings;
        char unknown0[4];
        char unknown1[24];
        char unknown2[4];
        char unknown3[24];
        char unknown4[4];
    };

    struct TRACE_CHAIN_ARRAY_ENTRY_17 {
        uint32 next_array_entry_index;
        uint32 total_block_load_count;
        uint32 unknown0;
        uint32 unknown1;
        uint32 unknown2;
    };

    struct FILE_METRICS_ARRAY_ENTRY_17 {
        uint32 start_time;
        uint32 duration;
        uint32 filename_string_offset;
        uint32 filename_string_number_of_characters;
        uint32 flags;
    };

    struct FILE_METRICS_ARRAY_ENTRY_23 {
        uint32 start_time;
        uint32 duration;
        uint32 average_duration;
        uint32 filename_string_offset;
        uint32 filename_string_number_of_characters;
        uint32 flags;
        uint64 ntfs_reference;
    };
    """
c_prefetch = cstruct().load(prefetch_def)

prefetch_version_structs = {
    17: (c_prefetch.FILE_INFORMATION_17, c_prefetch.FILE_METRICS_ARRAY_ENTRY_17),
    23: (c_prefetch.FILE_INFORMATION_23, c_prefetch.FILE_METRICS_ARRAY_ENTRY_23),
    30: (c_prefetch.FILE_INFORMATION_26, c_prefetch.FILE_METRICS_ARRAY_ENTRY_23),
    31: (c_prefetch.FILE_INFORMATION_26, c_prefetch.FILE_METRICS_ARRAY_ENTRY_23),
}


class Prefetch:
    def __init__(self, fh: BinaryIO):
        header_detect = c_prefetch.PREFETCH_HEADER_DETECT(fh.read(8))
        if header_detect.signature == b"MAM\x04":
            fh = BytesIO(lzxpress_huffman.decompress(fh))

        self.fh = fh
        self.fh.seek(0)
        self.header = c_prefetch.PREFETCH_HEADER(self.fh)
        self.version = self.identify()
        self.volumes = None
        self.metrics = None
        self.fn = None

        self.parse()

    def identify(self) -> int:
        self.fh.seek(0)
        version = self.header.version

        if version in prefetch_version_structs:
            return version

        raise NotImplementedError

    def parse(self) -> None:
        try:
            file_info_header, file_metrics_header = prefetch_version_structs.get(self.version)
            self.fh.seek(84)
            self.fn = file_info_header(self.fh)
            self.metrics = self.parse_metrics(metric_array_struct=file_metrics_header)
        except KeyError:
            raise NotImplementedError

    def parse_metrics(
        self, metric_array_struct: c_prefetch.FILE_METRICS_ARRAY_ENTRY_17 | c_prefetch.FILE_METRICS_ARRAY_ENTRY_23
    ) -> list[str | None]:
        metrics = []
        self.fh.seek(self.fn.metrics_array_offset)
        for _ in range(self.fn.number_of_file_metrics_entries):
            entry = metric_array_struct(self.fh)
            filename = self.read_filename(
                self.fn.filename_strings_offset + entry.filename_string_offset,
                entry.filename_string_number_of_characters,
            )
            metrics.append(filename.decode("utf-16-le"))
        return metrics

    def read_filename(self, off: int, size: int) -> bytes:
        offset = self.fh.tell()
        self.fh.seek(off)
        data = self.fh.read(size * 2)
        self.fh.seek(offset)  # reset pointer
        return data

    @property
    def latest_timestamp(self) -> datetime:
        """Get the latest execution timestamp inside the prefetch file."""
        return wintimestamp(self.fn.last_run_time)

    @property
    def previous_timestamps(self) -> list[datetime | None]:
        """Get the previous timestamps from the prefetch file."""
        try:
            # We check if timestamp actually has a value
            return [wintimestamp(timestamp) for timestamp in self.fn.last_run_remains if timestamp != 0]
        except AttributeError:
            # Header version 17 doesn't contain last_run_remains
            return []


class PrefetchPlugin(Plugin):
    """Windows prefetch plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.prefetchdir = self.target.resolve("%windir%/prefetch")

    def check_compatible(self) -> None:
        if len(list(self.prefetchdir.iterdir())) == 0:
            raise UnsupportedPluginError("No prefetch files found")

    @export(record=[PrefetchRecord, CompactPrefetchRecord])
    @arg("--compact", action="store_true", help="Compact the prefetch records")
    def prefetch(self, compact: bool = False) -> Iterator[PrefetchRecord | CompactPrefetchRecord]:
        """Return the content of all prefetch files.

        Prefetch is a memory management feature in Windows. It contains information (for example run count and
        timestamp) about executable applications that have been executed recently or are frequently executed.

        References:
            - https://www.geeksforgeeks.org/prefetch-files-in-windows/

        Yields PrefetchRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): Run timestamp.
            filename (path): The filename.
            prefetch (path): The prefetch entry.
            linkedfile (path): The linked file entry.
            runcount (int): The run count.

        with ``--compact``:

        Yields PrefetchRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): Run timestamp.
            filename (path): The filename.
            prefetch (path): The prefetch entry.
            linkedfiles (path[]): A list of linked files
            runcount (int): The run count.
            previousruns (datetime[]): Previous run non zero timestamps

        """
        for entry in self.prefetchdir.iterdir():
            if not entry.name.lower().endswith(".pf"):
                continue

            try:
                scca = Prefetch(entry.open())
            except Exception as e:
                self.target.log.warning("Failed to parse prefetch file: %s", entry)
                self.target.log.debug("", exc_info=e)
                continue

            filename = self.target.fs.path(scca.header.name.decode("utf-16-le", errors="ignore").split("\x00")[0])
            entry_name = self.target.fs.path(entry.name)

            if compact:
                yield CompactPrefetchRecord(
                    ts=scca.latest_timestamp,
                    filename=filename,
                    prefetch=entry_name,
                    linkedfiles=list(map(self.target.fs.path, scca.metrics)),
                    runcount=scca.fn.run_count,
                    previousruns=scca.previous_timestamps,
                    _target=self.target,
                )
            else:
                run_dates = [scca.latest_timestamp, *scca.previous_timestamps]
                for linked_file in scca.metrics:
                    for date in run_dates:
                        yield PrefetchRecord(
                            ts=date,
                            filename=filename,
                            prefetch=entry_name,
                            linkedfile=self.target.fs.path(linked_file),
                            runcount=scca.fn.run_count,
                            _target=self.target,
                        )



from __future__ import annotations

import csv
import gzip
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

AcquireOpenHandlesRecord = TargetRecordDescriptor(
    "filesystem/acquire_open_handles",
    [
        ("path", "name"),
        ("string", "handle_type"),
        ("string", "object"),
        ("varint", "unique_process_id"),
        ("varint", "handle_value"),
        ("varint", "granted_access"),
        ("varint", "creator_back_trace_index"),
        ("varint", "object_type_index"),
        ("varint", "handle_attributes"),
        ("varint", "reserved"),
    ],
)

AcquireHashRecord = TargetRecordDescriptor(
    "filesystem/acquire_hash",
    [
        ("path", "path"),
        ("filesize", "filesize"),
        ("digest", "digest"),
    ],
)


class AcquirePlugin(Plugin):
    """Returns records from data collected by Acquire."""

    __namespace__ = "acquire"

    def __init__(self, target: Target):
        super().__init__(target)
        self.hash_file = target.fs.path("$metadata$/file-hashes.csv.gz")
        self.open_handles_file = target.fs.path("$metadata$/open_handles.csv.gz")

    def check_compatible(self) -> None:
        if not self.hash_file.exists() and not self.open_handles_file.exists():
            raise UnsupportedPluginError("No hash file or open handles found")

    @export(record=AcquireHashRecord)
    def hashes(self) -> Iterator[AcquireHashRecord]:
        """Return file hashes collected by Acquire.

        An Acquire file container contains a file hashes csv when the hashes module was used. The content of this csv
        file is returned.
        """
        if self.hash_file.exists():
            with self.hash_file.open() as fh, gzip.open(fh, "rt") as gz_fh:
                for row in csv.DictReader(gz_fh):
                    yield AcquireHashRecord(
                        path=self.target.fs.path(row["path"]),
                        filesize=row["file-size"],
                        digest=(row["md5"] or None, row["sha1"] or None, row["sha256"] or None),
                        _target=self.target,
                    )

    @export(record=AcquireOpenHandlesRecord)
    def handles(self) -> Iterator[AcquireOpenHandlesRecord]:
        """Return open handles collected by Acquire.

        An Acquire file container contains an open handles csv when the handles module was used. The content of this csv
        file is returned.
        """
        if self.open_handles_file.exists():
            with self.open_handles_file.open() as fh, gzip.open(fh, "rt") as gz_fh:
                for row in csv.DictReader(gz_fh):
                    if name := row.get("name"):
                        row.update({"name": self.target.fs.path(name)})
                    yield AcquireOpenHandlesRecord(**row, _target=self.target)



"""Registry related abstractions."""

from __future__ import annotations

import fnmatch
import re
from collections import defaultdict
from functools import cached_property
from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO, NewType, TextIO

from dissect.regf import c_regf, regf

from dissect.target.exceptions import (
    RegistryError,
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
)
from dissect.target.helpers.utils import IntEnumMissing

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime
    from pathlib import Path

GLOB_INDEX_REGEX = re.compile(r"(^[^\\]*[*?[]|(?<=\\)[^\\]*[*?[])")
GLOB_MAGIC_REGEX = re.compile(r"[*?[]")

KeyType = regf.IndexLeaf | regf.FastLeaf | regf.HashLeaf | regf.IndexRoot | regf.KeyNode
"""The possible key types that can be returned from the registry."""

ValueType = int | str | bytes | list[str] | None
"""The possible value types that can be returned from the registry."""


class RegistryValueType(IntEnumMissing):
    """Registry value types as defined in ``winnt.h``.

    References:
        - https://learn.microsoft.com/en-us/windows/win32/sysinfo/registry-value-types
        - https://github.com/fox-it/dissect.regf/blob/main/dissect/regf/c_regf.py
    """

    NONE = c_regf.REG_NONE
    SZ = c_regf.REG_SZ
    EXPAND_SZ = c_regf.REG_EXPAND_SZ
    BINARY = c_regf.REG_BINARY
    DWORD = c_regf.REG_DWORD
    DWORD_BIG_ENDIAN = c_regf.REG_DWORD_BIG_ENDIAN
    LINK = c_regf.REG_LINK
    MULTI_SZ = c_regf.REG_MULTI_SZ
    RESOURCE_LIST = c_regf.REG_RESOURCE_LIST
    FULL_RESOURCE_DESCRIPTOR = c_regf.REG_FULL_RESOURCE_DESCRIPTOR
    RESOURCE_REQUIREMENTS_LIST = c_regf.REG_RESOURCE_REQUIREMENTS_LIST
    QWORD = c_regf.REG_QWORD


class RegistryHive:
    """Base class for registry hives."""

    def root(self) -> RegistryKey:
        """Return the root key of the hive."""
        return self.key("")

    def key(self, key: str) -> RegistryKey:
        """Retrieve a registry key from a specific path.

        Args:
            key: A path to a registry key within this hive.

        Raises:
            RegistryKeyNotFoundError: If the registry key could not be found.
        """
        raise NotImplementedError

    def keys(self, keys: str | list[str]) -> Iterator[RegistryKey]:
        """Retrieve all the registry keys in this hive from the given paths.

        Args:
            keys: A single path to find, or a list of paths to iterate over.
        """
        keys = [keys] if not isinstance(keys, list) else keys
        for key in keys:
            try:
                yield self.key(key)
            except RegistryError:  # noqa: PERF203
                pass


class RegistryKey:
    """Base class for registry keys.

    Args:
        hive: The registry hive to which this registry key belongs.
    """

    def __init__(self, hive: RegistryHive | None = None):
        self.hive = hive

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name}>"

    @property
    def ts(self) -> datetime:
        """Returns the last modified timestamp of this key."""
        return self.timestamp

    @property
    def name(self) -> str:
        """Returns the name of this key."""
        raise NotImplementedError

    @property
    def class_name(self) -> str:
        """Returns the class name of this key."""
        raise NotImplementedError

    @property
    def path(self) -> str:
        """Returns the path of this key."""
        raise NotImplementedError

    @property
    def timestamp(self) -> datetime:
        """Returns the last modified timestamp of this key."""
        raise NotImplementedError

    def get(self, key_path: str) -> RegistryKey:
        """Returns the RegistryKey pointed to by ``path``.

        Args:
            key_path: The path relative to this ``RegistryKey``.

        Returns:
            A relative ``RegistryKey``
        """
        key_path = key_path.strip("\\")
        if not key_path:
            return self

        key_path = f"{self.path}\\{key_path}"
        return self.hive.key(key_path)

    def subkey(self, subkey: str) -> RegistryKey:
        """Returns a specific subkey from this key.

        Args:
            subkey: The name of the subkey to retrieve.

        Raises:
            RegistryKeyNotFoundError: If this key has no subkey with the requested name.
        """
        raise NotImplementedError

    def subkeys(self) -> list[RegistryKey]:
        """Returns a list of subkeys from this key."""
        raise NotImplementedError

    Marker = NewType("Marker", object)
    __marker = Marker(object())

    def value(self, value: str, default: ValueType | Marker = __marker) -> RegistryValue:
        """Returns a specific value from this key.

        Args:
            value: The name of the value to retrieve.
            default: An optional argument that specifies the default return value in case the value cannot be found.

        Raises:
            RegistryValueNotFoundError: If this key has no value with the requested name and default is not defined.
        """
        try:
            return self._value(value)
        except RegistryValueNotFoundError:
            if default is self.__marker:
                raise
            return VirtualValue(self.hive, value, default)

    def _value(self, value: str) -> RegistryValue:
        raise NotImplementedError

    def values(self) -> list[RegistryValue]:
        """Returns a list of all the values from this key."""
        raise NotImplementedError


class RegistryValue:
    """Base class for registry values.

    Args:
        hive: The registry hive to which this registry value belongs.
    """

    def __init__(self, hive: RegistryHive | None = None):
        self.hive = hive

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name}={self.value!r}>"

    @property
    def name(self) -> str:
        """Returns the name of this value."""
        raise NotImplementedError

    @property
    def value(self) -> ValueType:
        """Returns the value of this value."""
        raise NotImplementedError

    @property
    def type(self) -> int:
        """Returns the type of this value.

        Reference:
            - https://docs.microsoft.com/en-us/windows/win32/sysinfo/registry-value-types
        """
        raise NotImplementedError


class VirtualHive(RegistryHive):
    """Virtual hive implementation."""

    def __init__(self):
        self._root = VirtualKey(self, "")

    def __repr__(self) -> str:
        return "<VirtualHive>"

    def make_keys(self, path: str) -> VirtualKey:
        """Create a key structure in this virtual hive from the given path.

        ``path`` must be a valid registry path to some arbitrary key in the registry. This method will traverse
        all the components of the path and create a key if it does not already exist.

        Example:
            The path ``test\\data\\something\\`` becomes::

                    "" <- root node
                    ├─ test
                    |  ├─ data
                    |  |  ├─ something

        Args:
            path: The registry path to create a key structure for.

        Returns:
            The :class:`VirtualKey` for the last path component.
        """
        path = path.strip("\\")
        key = self._root
        prev = None

        if not path:
            return key

        parts = path.split("\\")
        for i, part in enumerate(parts):
            try:
                prev = key
                key = key.subkey(part)
            except RegistryKeyNotFoundError:
                vkey = VirtualKey(self, "\\".join(parts[: i + 1]))
                prev.add_subkey(part, vkey)
                key = vkey

            # There was a bug with overlaying VirtualKeys on top of RegfKeys
            # In filesystem.py we solve this with layers, but this approach seems to work as well.
            # Basically, if we see that we're about to "overwrite" an existing key,
            # we copy it into a VirtualKey.
            if not isinstance(key, VirtualKey):
                vkey = VirtualKey(self, "\\".join(parts[: i + 1]))
                vkey.top = key
                prev.add_subkey(part, vkey)
                key = vkey

        return key

    def map_hive(self, path: str, hive: RegistryHive) -> None:
        """Map a different registry hive to a path in this registry hive.

        Future traversals to this path will continue from the root of the mapped hive.

        Args:
            path: The path at which to map the registry hive.
            hive: The hive to map to the path.
        """
        vkey = self.make_keys(path)
        vkey.top = hive.root()

    def map_key(self, path: str, key: RegistryKey) -> None:
        """Map an arbitrary :class:`RegistryKey` to a path in this hive.

        Args:
            path: The path at which to map the registry key.
            key: The :class:`RegistryKey` to map in this hive.
        """
        keypath, _, name = path.strip("\\").rpartition("\\")
        vkey = self.make_keys(keypath)
        vkey.add_subkey(name, key)

    def map_value(self, path: str, name: str, value: ValueType | RegistryValue) -> None:
        """Map an arbitrary value to a path and value name in this hive.

        Args:
            path: The path to the registry key that should hold the value.
            name: The name at which to store the value.
            value: The value to map to the specified location.
        """
        vkey = self.make_keys(path)
        vkey.add_value(name, value)

    def key(self, key: str) -> RegistryKey:
        path = key.strip("\\")
        vkey = self._root

        if not path:
            return vkey

        parts = path.split("\\")
        for part in parts:
            vkey = vkey.subkey(part)

        return vkey


class VirtualKey(RegistryKey):
    """Virtual key implementation."""

    def __init__(self, hive: RegistryHive, path: str, class_name: str | None = None):
        self._path = path
        if not path.strip("\\"):
            self._name = "VROOT"
        else:
            self._name = path.split("\\")[-1]
        self._class_name = class_name
        self._values: dict[str, RegistryValue] = {}
        self._subkeys: dict[str, RegistryKey] = {}
        self._timestamp: datetime = None
        self.top: RegistryKey = None
        super().__init__(hive=hive)

    def __contains__(self, key: str) -> bool:
        return key.lower() in self._subkeys

    def add_subkey(self, name: str, key: str) -> None:
        """Add a subkey to this key."""
        self._subkeys[name.lower()] = key

    def add_value(self, name: str, value: ValueType | RegistryValue) -> None:
        """Add a value to this key."""
        if not isinstance(value, RegistryValue):
            value = VirtualValue(self.hive, name, value)
        self._values[name.lower()] = value

    @property
    def name(self) -> str:
        return self._name

    @property
    def class_name(self) -> str:
        return self._class_name

    @property
    def path(self) -> str:
        return self._path

    @property
    def timestamp(self) -> datetime | None:
        if self.top:
            return self.top.timestamp

        if self._timestamp:
            return self._timestamp

        return None

    @timestamp.setter
    def timestamp(self, ts: datetime) -> None:
        self._timestamp = ts

    def subkey(self, subkey: str) -> RegistryKey:
        try:
            return self._subkeys[subkey.lower()]
        except KeyError:
            pass

        if self.top:
            try:
                return self.top.subkey(subkey)
            except RegistryKeyNotFoundError:
                pass

        raise RegistryKeyNotFoundError(subkey)

    def subkeys(self) -> list[RegistryKey]:  # Dict_values view
        res = {}

        for key in self._subkeys.values():
            res[key.name.lower()] = key

        if self.top:
            for key in self.top.subkeys():
                kn = key.name.lower()
                if kn in res:
                    continue
                res[kn] = key

        return res.values()

    def _value(self, value: str) -> RegistryValue:
        try:
            return self._values[value.lower()]
        except KeyError:
            pass

        if self.top:
            try:
                return self.top.value(value)
            except RegistryValueNotFoundError:
                pass
        raise RegistryValueNotFoundError(value)

    def values(self) -> list[RegistryValue]:
        res = {}

        for value in self._values.values():
            res[value.name.lower()] = value

        if self.top:
            for value in self.top.values():
                vn = value.name.lower()
                if vn in res:
                    continue
                res[vn] = value

        return res.values()


class VirtualValue(RegistryValue):
    """Virtual value implementation."""

    def __init__(self, hive: RegistryHive, name: str, value: ValueType):
        self._name = name
        self._value = value
        super().__init__(hive=hive)

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> ValueType:
        return self._value

    @property
    def type(self) -> RegistryValueType:
        return RegistryValueType.NONE


class HiveCollection(RegistryHive):
    """Hive implementation that is backed by multiple hives.

    The idea here is that you can open multiple version of the same
    hive (one regular, one with .LOG replayed and one RegBack).
    When opening a key, it would (try to) open it on every hive and
    return them in a KeyCollection.
    """

    def __init__(self, hives: list[RegistryHive] | None = None):
        self.hives = hives or []
        super().__init__()

    def __len__(self):
        return len(self.hives)

    def __iter__(self):
        return iter(self.hives)

    def __getitem__(self, index: int):
        return self.hives[index]

    def add(self, hive: RegistryHive) -> None:
        self.hives.append(hive)

    def key(self, key: str) -> KeyCollection:
        res = KeyCollection()

        for hive in self:
            try:
                res.add(hive.key(key))
            except RegistryKeyNotFoundError:  # noqa: PERF203
                pass

        if not res:
            raise RegistryKeyNotFoundError(key)

        return res

    def keys(self, keys: list | str) -> Iterator[RegistryKey]:
        keys = [keys] if not isinstance(keys, list) else keys
        for key in keys:
            try:
                yield from self.key(key)
            except RegistryError:  # noqa: PERF203
                pass

    def iterhives(self) -> Iterator[RegistryHive]:
        return iter(self.hives)


class KeyCollection(RegistryKey):
    """Key implementation that is backed by multiple keys.

    For example, both the current and the RegBack hive returned
    a key, but with different values. With a KeyCollection it's
    possible to iterate over all versions of this key.

    Things like traversing down subkeys works as expected, going
    down every key in it's collection.
    """

    def __init__(self, keys: list[RegistryKey] | None = None):
        self.keys = keys or []
        super().__init__()

    def __len__(self) -> int:
        return len(self.keys)

    def __iter__(self) -> Iterator[RegistryKey]:
        return iter(self.keys)

    def __getitem__(self, index: int) -> RegistryValue:
        return self.keys[index]

    def _key(self) -> RegistryKey:
        try:
            return self.keys[0]
        except IndexError:
            raise RegistryKeyNotFoundError

    def add(self, key: KeyCollection | RegistryKey) -> None:
        if isinstance(key, KeyCollection):
            self.keys.extend(key.keys)
        else:
            self.keys.append(key)

    @property
    def class_name(self) -> str:
        return self._key().class_name

    @property
    def name(self) -> str:
        return self._key().name

    @property
    def path(self) -> str:
        return self._key().path

    @property
    def timestamp(self) -> datetime:
        return self._key().timestamp

    def get(self, key_path: str) -> KeyCollection:
        ret = KeyCollection()
        for key in self:
            try:
                ret.add(key.get(key_path))
            except RegistryKeyNotFoundError:  # noqa: PERF203
                pass

        if not ret:
            raise RegistryKeyNotFoundError(key_path)

        return ret

    def subkey(self, subkey: str) -> KeyCollection:
        ret = KeyCollection()
        for key in self:
            try:
                ret.add(key.subkey(subkey))
            except RegistryKeyNotFoundError:  # noqa: PERF203
                pass

        if not ret:
            raise RegistryKeyNotFoundError(subkey)

        return ret

    def subkeys(self) -> list[KeyCollection]:
        ret = defaultdict(KeyCollection)
        for key in self:
            for sub in key.subkeys():
                ret[sub.name].add(sub)

        return ret.values()

    def _value(self, value: str) -> ValueCollection:
        ret = ValueCollection()
        for key in self:
            try:
                ret.add(key.value(value))
            except RegistryValueNotFoundError:  # noqa: PERF203
                pass

        if not ret:
            raise RegistryValueNotFoundError(value)

        return ret

    def values(self) -> list[ValueCollection]:
        ret = defaultdict(ValueCollection)
        for key in self:
            for value in key.values():
                ret[value.name].add(value)

        return ret.values()


class ValueCollection(RegistryValue):
    """Value implementation that is backed by multiple values.

    Same idea as KeyCollection, but for values.
    """

    def __init__(self, values: list[RegistryValue] | None = None):
        self.values = values or []
        super().__init__()

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        return iter(self.values)

    def _value(self) -> RegistryValue:
        try:
            return self.values[0]
        except IndexError:
            raise RegistryValueNotFoundError

    def add(self, value: RegistryValue) -> None:
        self.values.append(value)

    @property
    def name(self) -> str:
        return self._value().name

    @property
    def value(self) -> ValueType:
        return self._value().value

    @property
    def type(self) -> int:
        return self._value().type


class RegfHive(RegistryHive):
    """Registry implementation for regf hives."""

    def __init__(self, filepath: Path, fh: BinaryIO | None = None):
        fh = fh or filepath.open("rb")
        self.hive = regf.RegistryHive(fh)
        self.filepath = filepath

    def root(self) -> RegistryKey:
        return RegfKey(self, self.hive.root())

    def key(self, key: str) -> RegistryKey:
        try:
            return RegfKey(self, self.hive.open(key))
        except regf.RegistryKeyNotFoundError as e:
            raise RegistryKeyNotFoundError(key) from e


class RegfKey(RegistryKey):
    """Key implementation for regf keys."""

    def __init__(self, hive: RegistryHive, key: KeyType):
        self.key = key
        super().__init__(hive=hive)

    @property
    def name(self) -> str:
        return self.key.name

    @property
    def class_name(self) -> str:
        return self.key.class_name

    @property
    def path(self) -> str:
        return self.key.path

    @property
    def timestamp(self) -> datetime:
        return self.key.timestamp

    def subkey(self, subkey: str) -> RegistryKey:
        try:
            return RegfKey(self.hive, self.key.subkey(subkey))
        except regf.RegistryKeyNotFoundError as e:
            raise RegistryKeyNotFoundError(subkey) from e

    def subkeys(self) -> list[RegistryKey]:
        return [RegfKey(self.hive, k) for k in self.key.subkeys()]

    def _value(self, value: str) -> RegistryValue:
        try:
            return RegfValue(self.hive, self.key.value(value))
        except regf.RegistryValueNotFoundError as e:
            raise RegistryValueNotFoundError(value) from e

    def values(self) -> list[RegistryValue]:
        return [RegfValue(self.hive, v) for v in self.key.values()]


class RegfValue(RegistryValue):
    """Value implementation for regf values."""

    def __init__(self, hive: RegistryHive, kv: RegistryValue):
        self.kv = kv
        super().__init__(hive=hive)

    @property
    def name(self) -> str:
        return self.kv.name

    @property
    def value(self) -> ValueType:
        return self.kv.value

    @property
    def type(self) -> RegistryValueType:
        return RegistryValueType(self.kv.type)


class RegFlex:
    """A parser for text registry dumps (.reg files)."""

    def __init__(self):
        self.hives: dict[str, RegFlexHive] = {}

    def map_definition(self, fh: TextIO) -> None:
        """Parse a text registry export to a hive with keys and values.

        Args:
            fh: A file-like object opened in text mode of the registry export to parse.
        """
        vkey: RegFlexKey = None
        vhive: RegFlexHive = None

        for line in fh:
            line = line.strip()
            if not line:
                continue

            if line.startswith("[") and line.endswith("]"):
                if vkey:
                    vhive.map_key(vkey.path, vkey)

                hive, _, path = line[1:-1].partition("\\")
                hive = hive.upper()
                if hive not in self.hives:
                    self.hives[hive] = RegFlexHive()

                vhive = self.hives[hive]
                vkey = RegFlexKey(vhive, path)
                continue

            if line.startswith('"'):
                name, _, value = line.partition("=")
                name = name.strip('"')

                if value.endswith("\\"):
                    value = [value[:-1]]
                    while True:
                        next_line = next(fh).strip()
                        if next_line.endswith("\\"):
                            value.append(next_line[:-1])
                        else:
                            value.append(next_line)
                            break
                    value = "".join(value)

                vkey.add_value(name, RegFlexValue(vhive, name, value))

        vhive.map_key(vkey.path, vkey)


class RegFlexHive(VirtualHive):
    pass


class RegFlexKey(VirtualKey):
    pass


class RegFlexValue(VirtualValue):
    def __init__(self, hive: RegistryHive, name: str, value: ValueType):
        super().__init__(hive, name, value)

    @cached_property
    def _parse(self) -> tuple[RegistryValueType, ValueType]:
        return parse_flex_value(self._value)

    @property
    def value(self) -> ValueType:
        return self._parse[1]

    @property
    def type(self) -> RegistryValueType:
        return self._parse[0]


def parse_flex_value(value: str) -> tuple[RegistryValueType, ValueType]:
    """Parse values from text registry exports.

    Args:
        value: The value to parse.

    Raises:
        NotImplementedError: If ``value`` is not of a supported type for parsing.
    """
    if value.startswith('"'):
        return RegistryValueType.SZ, value.strip('"')

    vtype, _, value = value.partition(":")
    if vtype == "dword":
        return RegistryValueType.DWORD, int.from_bytes(bytes.fromhex(value), "big", signed=True)
    if "hex" in vtype:
        value = bytes.fromhex(value.replace(",", ""))
        if vtype == "hex":
            return RegistryValueType.BINARY, value

        # hex(T)
        # These values match regf type values
        vtype = int(vtype[4:5], 16)
        if vtype == regf.REG_NONE:
            decoded = value if value else None
        elif vtype == regf.REG_SZ or vtype == regf.REG_EXPAND_SZ:
            decoded = regf.try_decode_sz(value)
        elif vtype == regf.REG_BINARY:
            decoded = value
        elif vtype == regf.REG_DWORD:
            decoded = int.from_bytes(value, "little", signed=False)
        elif vtype == regf.REG_DWORD_BIG_ENDIAN:
            decoded = int.from_bytes(value, "big", signed=False)
        elif vtype == regf.REG_MULTI_SZ:
            d = BytesIO(value)

            r = []
            while d.tell() < len(value):
                s = regf.read_null_terminated_wstring(d)
                if s == "":
                    break

                r.append(s)

            decoded = r
        elif vtype == regf.REG_QWORD:
            decoded = int.from_bytes(value, "big", signed=False)
        else:
            raise NotImplementedError(f"Registry flex value type {vtype}")
        return RegistryValueType(vtype), decoded

    raise ValueError(f"Unsupported registry flex value type: {value!r}")


def has_glob_magic(pattern: str) -> bool:
    """Return whether ``pattern`` contains any glob patterns

    Args:
        pattern: The string to check on glob patterns.

    Returns:
        Whether ``pattern`` contains any glob patterns.
    """
    return GLOB_MAGIC_REGEX.search(pattern) is not None


def glob_split(pattern: str) -> tuple[str]:
    """Split a key path with glob patterns on the first key path part with glob patterns

    Args:
        pattern: A key path with glob patterns to split.

    Returns:
        A tuple of two strings, where the first contains the first number of
        key path parts (if any) which don't have a glob pattern. The second
        contains the rest of the key path with parts containing glob patterns.
    """
    first_glob = GLOB_INDEX_REGEX.search(pattern)

    if not first_glob:
        return pattern, ""

    pos = first_glob.start()
    return pattern[:pos], pattern[pos:]


def glob_ext(key_collection: KeyCollection, pattern: str) -> Iterator[KeyCollection]:
    """Yield all subkeys of ``key_collection`` that match the glob ``pattern``

    Args:
        key_collection: The ``KeyCollection`` to start the path pattern glob matching on.
        pattern: A key path with glob patterns.

    Yields:
        All subkeys that match ``pattern``
    """
    # This function operates by recursively stripping the last key path part
    # from pattern and recursively resolving the remaining part.

    if not has_glob_magic(pattern):
        # Short-cut if the pattern does not contain any actual globs.
        yield from glob_ext0(key_collection, pattern)
        return

    # The path pattern is stripped one key path part at a time for as long as
    # key_path contains glob patterns, starting at the end of the path.
    # At the final strip, key_path will be empty an last_key will contain the
    # very first part of the path pattern.
    key_path, _, last_key = pattern.strip("\\").rpartition("\\")

    if not key_path:
        # The end of the path is reached and the last key path part has glob
        # patterns (otherwise the shortcut at the start of the function would
        # have been taken).
        yield from glob_ext1(key_collection, last_key)
        return

    if has_glob_magic(key_path):
        # strip the next last key path part
        key_collections = glob_ext(key_collection, key_path)
    else:
        # This condition will generally not be true when the caller uses
        # glob_split() before calling this function (as
        # RegistryPlugin.glob_ext() does). In that case pattern (and thus
        # key_path), will start with a globbed path part.
        #
        # But if it is true, the key_path will not have any glob patterns, as
        # they were stripped off in the last path part (in last_key). So as a
        # short-cut the remaining key_path can be resolved directly instead of
        # recursively.
        key_collections = glob_ext0(key_collection, key_path)

    glob_in_key_path = glob_ext1 if has_glob_magic(last_key) else glob_ext0

    for key_collection in key_collections:
        yield from glob_in_key_path(key_collection, last_key)


def glob_ext0(key_collection: KeyCollection, key_path: str) -> Iterator[KeyCollection]:
    """Yield the subkey given by ``key_path`` relative to ``key_collection``

    Args:
        key_collection: The ``KeyCollection`` to yield the subkey from.
        key_path: The key path to the subkey, relative to ``key_collection``.

    Yields:
        The subkey from ``key_collection`` pointed to by ``key_path``.
    """
    try:
        yield key_collection.get(key_path)
    except RegistryKeyNotFoundError:
        pass


def glob_ext1(key_collection: KeyCollection, pattern: str) -> Iterator[KeyCollection]:
    """Yield all subkeys from ``key_collection`` which match the glob pattern ``pattern``

    Args:
        key_collection: The ``KeyCollection`` from which subkeys should be matched.
        pattern: The pattern a subkey must match.

    Yields:
        All KeyCollections of subkeys that match ``pattern``.
    """
    subkeys = key_collection.subkeys()
    # Note that the Windows registry is case insensitive, while
    # fnmatch.fnmatch() may not be on certain systems.
    pattern = pattern.lower()

    for subkey in subkeys:
        if fnmatch.fnmatch(subkey.name.lower(), pattern):
            yield subkey



from __future__ import annotations

import plistlib
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

AccountPolicyRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "macos/account_policy",
    [
        ("string", "generateduid"),
        ("datetime", "creation_time"),
        ("datetime", "password_last_time"),
        ("datetime", "failed_login_time"),
        ("varint", "failed_login_count"),
        ("path", "source"),
    ],
)


class UserPlugin(Plugin):
    """macOS user plugin."""

    # TODO: Parse additional user data like: HeimdalSRPKey, KerberosKeys, ShadowHashData, LinkedIdentity,
    # inputSources, smartCardSecureTokenData, smartCardSecureTokenUUID, unlockOptions, smb_sid

    USER_PATH = "/var/db/dslocal/nodes/Default/users"

    def check_compatible(self) -> None:
        if not self.target.fs.path(self.USER_PATH).exists():
            raise UnsupportedPluginError("No user directories found")

    @export(record=AccountPolicyRecord)
    def account_policy(self) -> Iterator[AccountPolicyRecord]:
        """Yield user account policy information."""

        # The data is not retrieved from the home folder of the user
        for user_details in self.target.user_details.all():
            user = plistlib.load(self.target.fs.path(user_details.user.source).open())

            if user.get("accountPolicyData"):
                generateduid = user.get("generateduid", [None])[0]

                for account_policy in user.get("accountPolicyData", []):
                    account_policy = plistlib.loads(account_policy)

                    yield AccountPolicyRecord(
                        generateduid=generateduid,
                        creation_time=account_policy.get("creationTime"),
                        password_last_time=account_policy.get("passwordLastSetTime"),
                        failed_login_count=account_policy.get("failedLoginCount"),
                        failed_login_time=account_policy.get("failedLoginTimestamp"),
                        source=self.target.fs.path(user_details.user.source),
                        _user=user_details.user,
                        _target=self.target,
                    )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.cstruct import cstruct

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

recent_files_def = """
    struct header {
        uint32  magic;
        uint32  unk0;
        uint32  unk1;
        uint32  unk2;
        uint32  checksum;
    };

    struct entry {
        uint32  length;
        wchar   path[length + 1];
    };
    """
c_recent_files = cstruct().load(recent_files_def)

RecentFileCacheRecord = TargetRecordDescriptor(
    "windows/recentfilecache",
    [
        ("path", "path"),
    ],
)


class RecentFileCachePlugin(Plugin):
    """Plugin that parses the RecentFileCache.bcf file."""

    def __init__(self, target: Target):
        super().__init__(target)
        self._recentfiles = self.target.resolve("%windir%/appcompat/programs/RecentFileCache.bcf")

    def check_compatible(self) -> None:
        if not self._recentfiles.exists():
            raise UnsupportedPluginError("Could not load RecentFileCache.bcf")

    @export(record=RecentFileCacheRecord)
    def recentfilecache(self) -> Iterator[RecentFileCacheRecord]:
        """Parse RecentFileCache.bcf.

        Yields RecentFileCacheRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            path (uri): The parsed path.
        """
        fh = self._recentfiles.open()

        c_recent_files.header(fh)
        while True:
            try:
                entry = c_recent_files.entry(fh)
                entry.path = entry.path.rstrip("\x00")

                yield RecentFileCacheRecord(
                    path=self.target.fs.path(entry.path),
                    _target=self.target,
                )
            except EOFError:  # noqa: PERF203
                break



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.cstruct import cstruct

from dissect.target.exceptions import RegistryKeyNotFoundError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.os.windows.credential.credential import WindowsCredentialPlugin
from dissect.target.plugins.os.windows.lsa import LSAPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

defaultpassword_def = """
struct DefaultPassword {
    DWORD   length;
    char    flags[4*3];
    WCHAR   data[length/2];
    // char    checksum_or_guid[0x10];
};
"""

c_defaultpassword = cstruct().load(defaultpassword_def)

DefaultPasswordRecord = TargetRecordDescriptor(
    "windows/credential/default_password",
    [
        ("datetime", "ts_mtime"),
        ("string", "default_password"),
        ("path", "source"),
    ],
)


class DefaultPasswordPlugin(WindowsCredentialPlugin):
    """Windows LSA DefaultPassword plugin."""

    __namespace__ = "defaultpassword"

    def check_compatible(self) -> None:
        if not self.target.has_function("lsa"):
            raise UnsupportedPluginError("LSA plugin not available on target")

    @export(record=DefaultPasswordRecord)
    def defaultpassword(self) -> Iterator[DefaultPasswordRecord]:
        """Yield decrypted Windows LSA DefaultPassword records.

        Extracts decrypted ``DefaultPassword`` values from the LSA.

        References:
            - https://learn.microsoft.com/en-us/troubleshoot/windows-server/user-profiles-and-logon/turn-on-automatic-logon
        """

        # Search for DefaultPassword values in the LSA
        for secret in ["DefaultPassword", "DefaultPassword_OldVal"]:
            if default_pass := self.target.lsa._secrets.get(secret):
                # Ignore irrelevant DefaultPassword references
                if "TBAL_{".encode("utf-16-le") in default_pass:
                    continue

                ts = None
                try:
                    ts = self.target.registry.key(
                        f"{LSAPlugin.SECURITY_POLICY_KEY}\\Secrets\\{secret.replace('_OldVal', '')}"
                    ).ts
                except RegistryKeyNotFoundError:
                    pass

                if default_pass.startswith(b"\x00" * 16) and len(default_pass) == 32:
                    yield DefaultPasswordRecord(
                        ts_mtime=ts,
                        default_password=default_pass[16:32].hex(),
                        source=f"HKLM\\SECURITY\\Policy\\Secrets\\{secret}",
                        _target=self.target,
                    )
                    continue

                try:
                    value = c_defaultpassword.DefaultPassword(default_pass).data
                except Exception as e:
                    self.target.log.warning("Failed to parse LSA %s value (%r): %s", secret, default_pass, e)
                    self.target.log.debug("", exc_info=e)
                    continue

                yield DefaultPasswordRecord(
                    ts_mtime=ts,
                    default_password=value,
                    source=f"HKLM\\SECURITY\\Policy\\Secrets\\{secret}",
                    _target=self.target,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.child.docker import DockerChildTargetPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_docker(target_linux_docker: Target) -> None:
    target_linux_docker.add_plugin(DockerChildTargetPlugin)
    children = sorted([child for _, child in target_linux_docker.list_children()], key=lambda r: r.path)

    assert len(children) == 3
    assert children[0].type == "docker"
    assert children[0].name == "epic_lichterman"

    assert [c.path for c in children] == [
        "/var/lib/docker/image/overlay2/layerdb/mounts/01b646bc043eb4ad72f3a64b4ffd9be2cbeb399e0a07497d749d724460ccad3a",
        "/var/lib/docker/image/overlay2/layerdb/mounts/589135d12011921ac6ce69753569da5f206f4bc792a9133727ddae860997ee66",
        "/var/lib/docker/image/overlay2/layerdb/mounts/f988f88e221d97930a665712cf16ab520f7e2c5af395660c145df93aebedf071",
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.database.ese import ESE
from flow.record.fieldtypes import digest

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.helpers.utils import findall
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

try:
    from asn1crypto.cms import ContentInfo
    from asn1crypto.core import Sequence

    HAS_ASN1 = True
except ImportError:
    HAS_ASN1 = False

HINT_NEEDLE = b"\x1e\x08\x00H\x00i\x00n\x00t"
PACKAGE_NAME_NEEDLE = b"\x06\n+\x06\x01\x04\x01\x827\x0c\x02\x01"
DIGEST_NEEDLES = {
    "md5": b"\x06\x08\x2a\x86\x48\x86\xf7\x0d\x02\x05",
    "sha1": b"\x06\x05\x2b\x0e\x03\x02\x1a",
    "sha_generic": b"\x06\x09\x08\x86\x48\x01\x65\x03\x04\x02",
    "sha256": b"\x06\t`\x86H\x01e\x03\x04\x02\x01\x05\x00",
}


CatrootRecord = TargetRecordDescriptor(
    "windows/catroot",
    [
        ("digest", "digest"),
        ("string[]", "hints"),
        ("string", "catroot_name"),
        ("path", "source"),
    ],
)


def _get_package_name(sequence: Sequence) -> str:
    """Parse sequences within a sequence and return the 'PackageName' value if it exists."""
    for value in sequence.native.values():
        # Value is an ordered dict that contains a sequence on index 1
        inner_sequence = Sequence.load(value.get("1"))
        # Key value is stored at index 0, value at index 2
        if "PackageName" in inner_sequence[0].native:
            return inner_sequence[2].native.decode("utf-16-le").strip("\x00")
    return None


def find_package_name(hint_buf: bytes) -> str | None:
    """Find a sequence that contains the 'PackageName' key and return the value if present."""
    for hint_offset in findall(hint_buf, PACKAGE_NAME_NEEDLE):
        # 7, 6 or 5 bytes before the package_name needle, a sequence starts (starts with b"0\x82" or b"0\x81").
        for sequence_needle in [b"0\x82", b"0\x81"]:
            if (sequence_offset := hint_buf.find(sequence_needle, hint_offset - 8, hint_offset)) == -1:
                continue

            hint_sequence = Sequence.load(hint_buf[sequence_offset:])
            return _get_package_name(hint_sequence)
    return None


class CatrootPlugin(Plugin):
    """Catroot plugin.

    Parses catroot files for hashes and file hints.
    """

    __namespace__ = "catroot"

    def __init__(self, target: Target):
        super().__init__(target)
        self.catroot_dir = self.target.resolve("%windir%/system32/catroot")
        self.catroot2_dir = self.target.resolve("%windir%/system32/catroot2")

    def check_compatible(self) -> None:
        if not HAS_ASN1:
            raise UnsupportedPluginError("Missing asn1crypto dependency")

        if next(self.catroot2_dir.rglob("catdb"), None) is None and next(self.catroot_dir.rglob("*.cat"), None) is None:
            raise UnsupportedPluginError("No catroot files or catroot ESE databases found")

    @export(record=CatrootRecord)
    def files(self) -> Iterator[CatrootRecord]:
        """Return the content of the catalog files in the CatRoot folder.

        A catalog file contains a collection of cryptographic hashes, or thumbprints. These files are generally used to
        verify the integrity of Windows operating system files, instead of per-file authenticode signatures.

        At the moment, parsing catalog files is done on best effort. ``asn1crypto`` is not able to fully parse the
        ``encap_content_info``, highly likely because Microsoft uses its own format. Future research should result in
        a more resilient and complete implementation of the ``catroot.files`` plugin.

        References:
            - https://www.thewindowsclub.com/catroot-catroot2-folder-reset-windows
            - https://docs.microsoft.com/en-us/windows-hardware/drivers/install/catalog-files

        Yields CatrootRecords with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            digest (digest): The parsed digest.
            hints (string[]): File hints, if present.
            catroot_name (string): Catroot name.
            source (path): Source of the catroot record.
        """
        # As far as known, Microsoft uses its own implementation to store the digest in the
        # encap_content_info along with an optional file hint. Here we parse the digest values
        # ourselves by looking for the corresponding digest needles in the raw encap_content_info
        # data. Furthermore, we try to find the file hint if it is present in that same raw data.
        for file in self.catroot_dir.rglob("*.cat"):
            if not file.is_file():
                continue

            try:
                buf = file.read_bytes()

                # TODO: Parse other data in the content info
                content_info = ContentInfo.load(buf)["content"]

                digest_type = content_info["digest_algorithms"].native[0].get("algorithm")
                encap_contents = content_info["encap_content_info"].contents
                needle = DIGEST_NEEDLES[digest_type]

                digests = []
                offset = None
                for offset in findall(encap_contents, needle):
                    # 4 bytes before the digest type, a sequence starts
                    objseq = Sequence.load(encap_contents[offset - 4 :])
                    # The second entry in the sequence is the digest string
                    raw_digest = objseq[1].native
                    hexdigest = raw_digest.hex()

                    file_digest = digest()
                    if len(hexdigest) == 32:
                        file_digest.md5 = hexdigest
                    elif len(hexdigest) == 40:
                        file_digest.sha1 = hexdigest
                    elif len(hexdigest) == 64:
                        file_digest.sha256 = hexdigest

                    digests.append(file_digest)

                # Finding the hint in encap_content_info is on best effort. In most of the cases,
                # there is a key "PackageName" available. We first try to parse the corresponding
                # value if it is present. If this does not succeed, we might be dealing with catroot
                # files containing the "hint" needle.
                # If both methods do not result in a file hint, there is either no hint available or
                # the format is not yet known and therefore not supported.
                hints = []
                try:
                    if offset:
                        # As far as known, the PackageName data is only present in the encap_content_info
                        # after the last digest.
                        hint_buf = encap_contents[offset + len(needle) + len(raw_digest) + 2 :]

                        # First try to find to find the "PackageName" value, if it's present.
                        hint = find_package_name(hint_buf)
                        if hint:
                            hints.append(hint)

                    # If the package_name needle is not found or it's not present in the first 7 bytes of the hint_buf
                    # We are probably dealing with a catroot file that contains "hint" needles.
                    if not hints:
                        for hint_offset in findall(encap_contents, HINT_NEEDLE):
                            # Either 3 or 4 bytes before the needle, a sequence starts
                            bytes_before_needle = 3 if encap_contents[hint_offset - 3] == 48 else 4
                            name_sequence = Sequence.load(encap_contents[hint_offset - bytes_before_needle :])

                            hint = name_sequence[2].native.decode("utf-16-le").strip("\x00")
                            hints.append(hint)

                except Exception as e:
                    self.target.log.debug("", exc_info=e)

                # Currently, it is not known how the file hints are related to the digests. Therefore, each digest
                # is yielded as a record with all of the file hints found.
                # TODO: find the correlation between the file hints and the digests in catroot files.
                for file_digest in digests:
                    yield CatrootRecord(
                        digest=file_digest,
                        hints=hints,
                        catroot_name=file.name,
                        source=file,
                        _target=self.target,
                    )

            except Exception as e:
                self.target.log.error("An error occurred while parsing the catroot file %s: %s", file, e)  # noqa: TRY400
                self.target.log.debug("", exc_info=e)

    @export(record=CatrootRecord)
    def catdb(self) -> Iterator[CatrootRecord]:
        """Return the hash values present in the catdb files in the catroot2 folder.

        The catdb file is an ESE database file that contains the digests of the catalog files present on the system.
        This database is used to speed up the process of validating a Portable Executable (PE) file.

        Note: catalog files can include file hints, however these seem not to be present in the catdb files.

        References:
            - https://www.thewindowsclub.com/catroot-catroot2-folder-reset-windows
            - https://docs.microsoft.com/en-us/windows-hardware/drivers/install/catalog-files

        Yields CatrootRecords with the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            digest (digest): The parsed digest.
            hints (string[]): File hints, if present.
            catroot_name (string): Catroot name.
            source (path): Source of the catroot record.
        """
        for ese_file in self.catroot2_dir.rglob("catdb"):
            with ese_file.open("rb") as fh:
                ese_db = ESE(fh)

                for hash_type, table_name in [("sha256", "HashCatNameTableSHA256"), ("sha1", "HashCatNameTableSHA1")]:
                    try:
                        table = ese_db.table(table_name)
                    except KeyError as e:
                        self.target.log.warning("EseDB %s has no table %s", ese_file, table_name)
                        self.target.log.debug("", exc_info=e)
                        continue

                    for record in table.records():
                        file_digest = digest()

                        try:
                            setattr(file_digest, hash_type, record.get("HashCatNameTable_HashCol").hex())
                            catroot_names = record.get("HashCatNameTable_CatNameCol").decode().rstrip("|").split("|")
                        except Exception as e:
                            self.target.log.warning("Unable to parse catroot names for %s in %s", record, ese_file)
                            self.target.log.debug("", exc_info=e)
                            continue

                        for catroot_name in catroot_names:
                            yield CatrootRecord(
                                digest=file_digest,
                                hints=None,
                                catroot_name=catroot_name,
                                source=ese_file,
                                _target=self.target,
                            )



from __future__ import annotations

import sys
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.helpers.logging import get_logger
from dissect.target.loader import Loader
from dissect.target.plugins.os.default._os import DefaultOSPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

log = get_logger(__name__)


class DirectLoader(Loader):
    def __init__(self, paths: list[str | Path], case_sensitive: bool = False):
        self.case_sensitive = case_sensitive
        self.paths = [(Path(path) if not isinstance(path, Path) else path).resolve() for path in paths]

    def __repr__(self) -> str:
        # We need to override __repr__ because we do not define self.path
        return f"{self.__class__.__name__}({str(self.paths)!r})"

    @staticmethod
    def detect(path: Path) -> bool:
        return False

    def map(self, target: Target) -> None:
        if not self.case_sensitive and self.check_case_insensitive_overlap():
            log.warning(
                "Direct mode used in case insensitive mode, but this will cause files overlap, "
                "consider using --direct-sensitive"
            )
        vfs = VirtualFilesystem(case_sensitive=self.case_sensitive)
        for path in self.paths:
            if path.is_file():
                vfs.map_file(str(path), path)
            elif path.is_dir():
                vfs.map_dir(str(path), path)

        target.filesystems.add(vfs)
        target._os_plugin = DefaultOSPlugin

    def check_case_insensitive_overlap(self) -> bool:
        """Verify if two differents files will have the same path in a case-insensitive scenario."""
        if sys.version_info >= (3, 12) or sys.platform != "win32":

            def get_files(path: Path) -> Iterator[Path]:
                if not path.exists():
                    return
                if path.is_file():
                    yield path
                # Recursively find all files in the directory
                yield from path.rglob("*")
        else:

            def get_files(path: Path, max_depth: int = 7) -> Iterator[Path]:
                # rglob seems to have issue on windows with python <3.12 when working on a case sensitive FS. Thus
                # we use another implementation without using rglob.
                # Probably related to https://github.com/python/cpython/issues/94537

                if max_depth == 0:
                    return
                if not path.exists():
                    return
                if path.is_dir():
                    for f in path.iterdir():
                        if f.is_dir():
                            yield from get_files(f, max_depth=max_depth - 1)
                        else:
                            yield f
                elif path.is_file():
                    yield path

        # Create a flat list of all file paths from all input directories
        all_paths = chain.from_iterable(get_files(p) for p in self.paths)
        # Filter out directories, keeping only files
        all_files = [p for p in all_paths if p.is_file()]
        # Compare the count of all files with the count of unique, lowercased file paths

        return len({str(p).lower() for p in all_files}) != len(all_files)



from __future__ import annotations

import importlib
import importlib.resources
from locale import normalize
from pathlib import Path

import defusedxml.ElementTree as ET


def normalize_timezone(input: str) -> str:
    """Returns normalized timezone format per IANA TZ standard.

    Takes a Windows registry ``TimeZoneKeyName`` string as input and translates it to IANA TZ format.
    Will return the IANA preferred territory ``001`` value of the corresponding timezone.

    For example, ``Pacific Standard Time`` is translated to ``America/Los_Angeles``.
    """

    return WINDOWS_ZONE_MAP.get(input, input)


def normalize_language(input: str) -> str:
    """Returns normalized locales per ISO-3166. Takes Unix LANG locales and Windows registry languages as input.

    Output will be in the format ``ISO-3166-1-alpha-2-code_ISO-3166-2``, e.g.: ``en_US``, ``nl_NL`` or ``en_GB``.

    References:
        - https://en.wikipedia.org/wiki/ISO_3166
        - https://en.wikipedia.org/wiki/ISO_3166-1
        - https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
        - https://en.wikipedia.org/wiki/ISO_3166-2
    """

    return normalize(input.replace("-", "_", 1)).split(".")[0]


def get_resource_string(path: str) -> str:
    return _get_resource_path(path).read_text()


def _get_resource_path(path: str) -> Path:
    root = importlib.resources.files(__package__) if __package__ else Path(__file__).parent
    fpath = root.joinpath(path)

    if not fpath.exists():
        raise IOError(f"Can't find resource {fpath}")

    return fpath


WINDOWS_ZONE_MAP = {
    child.attrib["other"]: child.attrib["type"]
    for child in ET.fromstring(get_resource_string("data/windowsZones.xml")).findall(
        "./windowsZones/mapTimezones/mapZone"
    )
    if child.attrib["territory"] == "001"
}

WINDOWS_ZONE_MAP["UTC"] = "UTC"  # Change 'Etc/UTC' to 'UTC' to be consistent across operating systems.



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.regf.userassist import UserAssistPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target

USERASSIST_DATA_VERSION0 = bytes.fromhex(
    "0100000000060000341ad99d3c98d808",
)
USERASSIST_DATA_VERSION3_8 = bytes.fromhex("03e08d0e01000000")
USERASSIST_DATA_VERSION3_16 = bytes.fromhex(
    "000000001a000000d8a599defebbd101",
)
USERASSIST_DATA_VERSION5 = bytes.fromhex(
    "ffffffff000000000000000000000000000080bf000080bf000080bf000080bf"
    "000080bf000080bf000080bf000080bf000080bf000080bfffffffff00000000"
    "0000000000000000"
)


def test_userassist_plugin(target_win_users: Target, hive_hku: VirtualHive) -> None:
    userassist_key_name = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\UserAssist"
    userassist_key = VirtualKey(hive_hku, userassist_key_name)

    version0_key = VirtualKey(hive_hku, "{F4E57C4B-2036-45F0-A9AB-443BCFE33D9F}")
    version0_count_key = VirtualKey(hive_hku, "Count")
    version0_key.add_subkey("Count", version0_count_key)
    version0_count_key.add_value(
        "Q:\\frghc64.rkr",
        VirtualValue(hive_hku, "Q:\\frghc64.rkr", USERASSIST_DATA_VERSION0),
    )

    version3_key = VirtualKey(hive_hku, "{75048700-EF1F-11D0-9888-006097DEACF9}")
    version3_key.add_value("Version", VirtualValue(hive_hku, "Version", 3))
    version3_count_key = VirtualKey(hive_hku, "Count")
    version3_key.add_subkey("Count", version3_count_key)
    version3_count_key.add_value(
        "Zvpebfbsg.Trgfgnegrq_8jrxlo3q8oojr!Ncc",
        VirtualValue(hive_hku, "Zvpebfbsg.Trgfgnegrq_8jrxlo3q8oojr!Ncc", USERASSIST_DATA_VERSION3_8),
    )
    version3_count_key.add_value(
        "HRZR_PGYFRFFVBA",
        VirtualValue(hive_hku, "HRZR_PGYFRFFVBA", USERASSIST_DATA_VERSION3_16),
    )

    version5_key = VirtualKey(hive_hku, "{CAA59E3C-4792-41A5-9909-6A6A8D32490E}")
    version5_key.add_value("Version", VirtualValue(hive_hku, "Version", 5))
    version5_count_key = VirtualKey(hive_hku, "Count")
    version5_key.add_subkey("Count", version5_count_key)
    version5_count_key.add_value(
        "HRZR_PGYPHNPbhag:pgbe",
        VirtualValue(hive_hku, "HRZR_PGYPHNPbhag:pgbe", USERASSIST_DATA_VERSION5),
    )

    userassist_key.add_subkey(version0_key.name, version0_key)
    userassist_key.add_subkey(version3_key.name, version3_key)
    userassist_key.add_subkey(version5_key.name, version5_key)
    hive_hku.map_key(userassist_key_name, userassist_key)

    target_win_users.add_plugin(UserAssistPlugin)

    results = list(target_win_users.userassist())

    assert len(results) == 4
    assert str(results[0].ts) == "1601-01-01 00:00:00+00:00"
    assert results[0].path == "D:\\setup64.exe"
    assert results[0].number_of_executions is None
    assert results[0].application_focus_count is None
    assert results[0].application_focus_duration is None
    assert str(results[1].ts) == "1601-01-01 00:00:00+00:00"
    assert results[1].path == "Microsoft.Getstarted_8wekyb3d8bbwe!App"
    assert results[1].number_of_executions is None
    assert results[1].application_focus_count is None
    assert results[1].application_focus_duration is None
    assert str(results[2].ts) == "2016-06-01 12:12:30.115376+00:00"
    assert results[2].path == "UEME_CTLSESSION"
    assert results[2].number_of_executions == 26
    assert results[2].application_focus_count is None
    assert results[2].application_focus_duration is None
    assert str(results[3].ts) == "1601-01-01 00:00:00+00:00"
    assert results[3].path == "UEME_CTLCUACount:ctor"
    assert results[3].number_of_executions == 0
    assert results[3].application_focus_count == 0
    assert results[3].application_focus_duration == 0



from __future__ import annotations

import io
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

from dissect.target.filesystem import VirtualFile, VirtualFilesystem
from dissect.target.tools.fs import _extract_path, cp
from dissect.target.tools.fs import main as target_fs

if TYPE_CHECKING:
    from pathlib import Path


@pytest.fixture
def vfs(files: list[str]) -> VirtualFilesystem:
    vfs = VirtualFilesystem()
    for file in files:
        if file[-1] == "/":
            vfs.makedirs(file)
        else:
            vfs.map_file_entry(file, VirtualFile(vfs, file, io.BytesIO()))
    return vfs


@pytest.mark.parametrize(
    ("path", "expected_files"),
    [
        ("root/.bash_history", 1),
        ("this/path/doesntexist", 0),
        ("**/*_history", 4),
    ],
)
def test_target_fs(
    path: str, expected_files: int, tmp_path: Path, capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-fs", "tests/_data/tools/info/image.tar", "cp", path, "-o", str(tmp_path)])

        target_fs()
        stdout, _ = capsys.readouterr()

        if expected_files > 0:
            lines = [line for line in stdout.split("\n") if line != ""]
            assert len(lines) == expected_files
        else:
            assert stdout == "[!] Path doesn't exist\n"


@pytest.mark.parametrize("files", [["file"]])
def test_extract_file(vfs: VirtualFilesystem, tmp_path: Path) -> None:
    output_path = tmp_path / "file"

    try:
        _extract_path(vfs.path("file"), output_path)
    except Exception:  # noqua
        # The files are virtual, so we expect the method to raise an exception
        pass

    assert output_path.exists()


@pytest.mark.parametrize("files", [[]])
def test_file_not_exist(vfs: VirtualFilesystem, tmp_path: Path) -> None:
    output_path = tmp_path / "file"

    _extract_path(vfs.path("file"), output_path)

    assert not output_path.exists()


@pytest.mark.parametrize("files", [["dir/"]])
def test_extract_directory(vfs: VirtualFilesystem, tmp_path: Path) -> None:
    output_path = tmp_path / "out"

    _extract_path(vfs.path("dir"), output_path)

    assert output_path.exists()


@pytest.mark.parametrize("files", [["dir/", "dir/test"]])
def test_cp_file_path(vfs: VirtualFilesystem, tmp_path: Path) -> None:
    output_path = tmp_path / "out"

    args = Mock()
    args.output = str(output_path)

    cp(None, vfs.path("dir/test"), args)

    assert output_path.joinpath("test").exists()


@pytest.mark.parametrize("files", [["dir/", "dir/test"]])
def test_cp_directory(vfs: VirtualFilesystem, tmp_path: Path) -> None:
    output_path = tmp_path / "out"

    args = Mock()
    args.output = str(output_path)

    cp(None, vfs.path("dir"), args)

    assert output_path.joinpath("test").exists()


@pytest.mark.parametrize("files", [[]])
def test_cp_non_existing_file(vfs: VirtualFilesystem, tmp_path: Path) -> None:
    output_path = tmp_path / "out"

    args = Mock()
    args.output = str(output_path)

    cp(None, vfs.path("dir/test"), args)

    assert not output_path.exists()


@pytest.mark.parametrize(
    "files",
    [["dir/", "dir/test", "dir/subdirectory_1/", "dir/subdirectory_2/", "dir/subdirectory_3/subdirectory_4/"]],
)
def test_cp_subdirectories(vfs: VirtualFilesystem, files: list[str], tmp_path: Path) -> None:
    output_path = tmp_path / "out"

    args = Mock()
    args.output = str(output_path)

    cp(None, vfs.path("dir/"), args)

    filesystem_files = (file.replace("dir/", "") for file in files)

    for directories in filesystem_files:
        assert output_path.joinpath(directories).exists()



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest

from dissect.target.plugins.os.windows.credential.defaultpassword import DefaultPasswordPlugin
from dissect.target.plugins.os.windows.lsa import LSAPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("reg_key", "reg_value", "expected_value"),
    [
        pytest.param(
            "DefaultPassword",
            bytes.fromhex("00000000000000000000000000000000d41d8cd98f00b204e9800998ecf8427e"),
            "d41d8cd98f00b204e9800998ecf8427e",
            id="small",
        ),
        pytest.param(
            "DefaultPassword_OldVal",
            bytes.fromhex(
                "10000000000000000000000000000000700061007300730077006f0072006400d41d8cd98f00b204e9800998ecf8427e"
            ),
            "password",
            id="regular",
        ),
        pytest.param(
            "DefaultPassword",
            bytes.fromhex("10000000000000000000000000000000700061007300730077006f0072006400"),
            "password",
            id="no-trailer",
        ),
    ],
)
def test_windows_credential_defaultpassword_lsa(
    target_win: Target, reg_key: str, reg_value: bytes, expected_value: str
) -> None:
    """Test if we can parse a DefaultPassword LSA entry."""

    with mock.patch(
        "dissect.target.plugins.os.windows.lsa.LSAPlugin._secrets",
        new_callable=mock.PropertyMock,
        return_value={reg_key: reg_value},
    ):
        target_win.add_plugin(LSAPlugin, check_compatible=False)
        target_win.add_plugin(DefaultPasswordPlugin)

        records = list(target_win.credential.defaultpassword())

        assert len(records) == 1

        assert records[0].default_password == expected_value
        assert records[0].source == f"HKLM\\SECURITY\\Policy\\Secrets\\{reg_key}"
